import logging

import jsonpath_rw
from jsonpointer import JsonPointer

from .stages import JsonStage, OffsetStage
from .utils import get_obj_id, init_stage_spec
from .wflownode import YadageNode

log = logging.getLogger(__name__)



class WorkflowView(object):
    '''
    Provides a 'view' of the overall workflow object that corresponds
    to a particular level of nesting. That is, this presents the scope
    within which extension rules operate (i.e. add steps, reference
    other steps, etc)
    '''

    def __init__(self, workflowobj, offset=''):
        self.wflow = workflowobj
        self.dag = workflowobj.dag
        self.rules = workflowobj.rules
        self.applied_rules = workflowobj.applied_rules
        self.offset = offset
        self.steps = JsonPointer(self.offset).resolve(workflowobj.stepsbystage)
        self.bookkeeper = JsonPointer(self.offset).resolve(workflowobj.bookkeeping)
        self.values = JsonPointer(self.offset).resolve(workflowobj.values)

    def view(self, offset):
        '''
        return a view with an additional offset to this view's offset
        '''
        return self.wflow.view(self._makeoffset(offset))

    def query(self, query, collection):
        '''
        :return
        '''
        matches = jsonpath_rw.parse(query).find(collection)
        return matches

    def getSteps(self, query):
        '''
        returns steps related to the JSONPath query.
        if a query points to a stage (say 'stagename'):
        will return all toplevel steps.. no recursion into subworkflows
        if a query points to steps (e.g. 'stagename[*]')
        will return a steps directly

        '''
        nodeids = []

        matches = self.query(query, self.steps)
        for match in matches:
            value = match.value
            #step endpoint case
            if isinstance(value,dict) and '_nodeid' in value:
                nodeids.append(value['_nodeid'])
            #stage endpoint case
            elif isinstance(value,list):
                for item in value:
                    if '_nodeid' in item:
                        nodeids.append(item['_nodeid'])

        result = [self.dag.getNode(nodeid) for nodeid in nodeids]
        return result

    def _makeoffset(self, offset):
        '''
        prepare a full offset based on this views' offset and a relative offset

        :param offset: the relative offset
        '''
        thisoffset = JsonPointer(offset)
        if self.offset:
            fulloffset = JsonPointer.from_parts(JsonPointer(self.offset).parts + thisoffset.parts).path
        else:
            fulloffset = thisoffset.path
        return fulloffset

    def getRule(self, name=None, offset='', identifier=None):
        '''retrieve a rule by offset or name or identifier'''
        fulloffset = self._makeoffset(offset)
        for x in self.rules + self.applied_rules:
            if x.identifier == identifier or (x.offset == fulloffset and x.rule.name == name):
                return x
        return None

    def init(self, initdata, init_provider = None, used_inputs = None, name='init', discover = False):
        '''
        initialize this scope by adding an initialization stage.

        :param inidata: initialization JSON data
        '''
        spec = init_stage_spec(initdata, discover, used_inputs or [], name)
        self.addRule(JsonStage(spec, init_provider), self.offset)

    def addRule(self, rule, offset=''):
        '''
        add a DAG extension rule, possibly with a scope offset
        '''
        thisoffset = JsonPointer(offset)
        offsetstage = OffsetStage(rule, self._makeoffset(offset))
        self.rules += [offsetstage]
        createOffsetMeta(thisoffset.path, self.bookkeeper)
        thisoffset.resolve(self.bookkeeper)['_meta']['stages'] += [offsetstage.identifier]
        return offsetstage.identifier

    def addValue(self, key, value):
        v = self.values.setdefault('_values', {})
        if key in v:
            raise RuntimeError('cannot overwrite value')
        v[key] = value

    def getValue(self, key):
        return self.values['_values'][key]

    def addStep(self, task, stage, depends_on=None):
        '''
        adds a node to the DAG connecting it to the passed depending nodes
        while tracking that it was added by the specified stage

        :param task: the task object for the step
        :param stage: the stage name
        :param depends_on: dependencies of this step
        '''

        self.steps.setdefault(stage,[])
        node = YadageNode(task.metadata['name'], task, identifier=get_obj_id(task))
        node.task.metadata['wflow_node_id'] = node.identifier
        node.task.metadata['wflow_offset'] = self.offset
        node.task.metadata['wflow_stage'] = stage
        node.task.metadata['wflow_stage_node_idx'] = len(self.steps[stage])
        node.task.metadata['wflow_hints'] = {'is_purepub': task.pubOnlyTask()}

        self.dag.addNode(node, depends_on=depends_on)
        self.steps[stage].append({'_nodeid': node.identifier})
        self.bookkeeper['_meta']['steps'] += [node.identifier]
        log.info('added node %s', node)
        return node

    def addWorkflow(self, rules, stage=None):
        '''
        add a (sub-)workflow (i.e. list of stages) to the overall workflow
        '''
        offset = ''
        if stage is not None:
            #make sure storage for the 'authoring' stage is present and
            #register the workflow as part of that 'author'
            #needed e.g. for predicate handlers trying to determing if
            #the author stage is done
            self.steps.setdefault(stage,[]).append({})
            offset = JsonPointer.from_parts([stage, len(self.steps[stage]) - 1]).path
            self.steps[stage][-1]['_offset'] = offset

            self.values.setdefault(stage,[]).append({})
            offset = JsonPointer.from_parts([stage, len(self.values[stage]) - 1]).path
            self.values[stage][-1] = {}

        for rule in rules:
            self.addRule(rule, offset)

def createOffsetMeta(offset, bookkeeping):
    '''
    sets up a location to track rule and step ids for a given scope offset
    '''
    pointer = JsonPointer(offset)
    view = bookkeeping
    for x in pointer.parts:
        view = view.setdefault(x,{})
    pointer.resolve(bookkeeping).setdefault('_meta',{'stages': [], 'steps': []})



#!/usr/bin/env python
import logging
import os

import click
import yaml

import yadage.steering_api as steering_api
import yadage.utils as utils
import yadageschemas
from packtivity.plugins import enable_plugins

log = logging.getLogger(__name__)

RC_FAILED = 1
RC_SUCCEEDED = 0


def from_file(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    data = {}
    verbosity = 'INFO'
    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    click.secho('running from file')
    for v in value:
        data.update(**yaml.load(v))
    data['backend']  = utils.setupbackend_fromstring(data.pop('backend'),data.pop('backendopts'))

    enable_plugins(data.pop('plugins'))

    rc = RC_FAILED
    try:
        steering_api.run_workflow(**data)
        rc = RC_SUCCEEDED
    except:
        log.exception('workflow failed')
    if rc != RC_SUCCEEDED:
        exc = click.exceptions.ClickException(
            click.style("Workflow failed", fg='red')
        )
        exc.exit_code = rc
        raise exc

    ctx.exit()

@click.command()
@click.argument('dataarg', default = 'work')
@click.argument('workflow', default = 'workflow.yml')
@click.argument('initfiles', nargs=-1)
@click.option('-b', '--backend', default='multiproc:auto', help = 'packtivity backend string')
@click.option('-c', '--cache', default='')
@click.option('-d', '--dataopt', multiple=True, default=None, help = 'options for the workflow data state')
@click.option('-e', '--schemadir', default=yadageschemas.schemadir, help = 'schema directory for workflow validation')
@click.option('-f', '--from-file', expose_value=False, multiple=True, type=click.File('rb'), help = 'read entire configuration from file, no other flags settings are read.', callback=from_file, is_eager = True)
@click.option('-g', '--strategy', help = 'set execution stragegy')
@click.option('-i', '--loginterval', default=30, help = 'adage tracking interval in seconds')
@click.option('-k', '--backendopt', multiple=True, default=None, help = 'options for the workflow data state')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-m', '--metadir', default=None, help = 'directory to store workflow metadata')
@click.option('-o', '--ctrlopt', multiple=True, default=None, help = 'options for the workflow controller')
@click.option('-p', '--parameter', multiple=True, help = '<parameter name>=<yaml string> input parameter specifcations ')
@click.option('-r', '--controller', default = 'frommodel', help = 'controller')
@click.option('-s', '--modelsetup', default='inmem', help = 'wflow state model')
@click.option('-t', '--toplevel', default=os.getcwd(), help = 'toplevel uri to be used to resolve workflow name and references from')
@click.option('-u', '--updateinterval', default=0.02, help = 'adage graph inspection interval in seconds')
@click.option('-v', '--verbosity', default='INFO', help = 'logging verbosity')
@click.option('--accept-metadir/--no-accept-metadir', default=False)
@click.option('--plugins', default=None)
@click.option('--validate/--no-validate', default=True, help = 'en-/disable workflow spec validation')
@click.option('--visualize/--no-visualize', default=False, help = 'visualize workflow graph')
def main(dataarg,
         workflow,
         initfiles,
         controller,
         ctrlopt,
         toplevel,
         verbosity,
         loginterval,
         updateinterval,
         schemadir,
         backend,
         dataopt,
         backendopt,
         strategy,
         modelsetup,
         modelopt,
         metadir,
         parameter,
         validate,
         visualize,
         cache,
         plugins,
         accept_metadir
         ):


    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    from packtivity.plugins import enable_plugins
    if plugins:
        enable_plugins(plugins.split(','))

    initdata    = utils.getinit_data(initfiles, parameter)
    dataopts    = utils.options_from_eqdelimstring(dataopt)
    backendopts = utils.options_from_eqdelimstring(backendopt)
    modelopts   = utils.options_from_eqdelimstring(modelopt)
    ctrlopts    = utils.options_from_eqdelimstring(ctrlopt)

    backend  = utils.setupbackend_fromstring(backend,backendopts)
    rc = RC_FAILED
    try:
        steering_api.run_workflow(
            workflow = workflow,
            toplevel = toplevel,
            validate = validate,
            schemadir = schemadir,

            initdata = initdata,
            controller = controller,
            ctrlopts = ctrlopts,
            backend = backend,
            cache = cache,

            dataarg = dataarg,
            dataopts = dataopts,
            metadir = metadir,
            accept_metadir = accept_metadir,
            modelsetup = modelsetup,
            modelopts = modelopts,
            updateinterval = updateinterval,
            loginterval = loginterval,
            visualize = visualize,
            strategy = strategy,
        )
        rc = RC_SUCCEEDED
    except:
        log.exception('workflow failed')
    if rc != RC_SUCCEEDED:
        exc = click.exceptions.ClickException(
            click.style("Workflow failed", fg='red')
        )
        exc.exit_code = rc
        raise exc

if __name__ == '__main__':
    main()



import copy
import hashlib
import importlib
import json
import logging
import os
import uuid
import zipfile

import click
import jq
import jsonpointer
import yaml

from yadageschemas.utils import WithJsonRefEncoder


class outputReference(object):

    def __init__(self, stepid, pointer):
        self.stepid = stepid
        self.pointer = pointer

    def __repr__(self):
        return 'outputReference {}#{}'.format(self.stepid, self.pointer.path)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        return cls(data['stepid'], jsonpointer.JsonPointer(data['pointer_path']))

    def json(self):
        return {
            'stepid': self.stepid,
            'pointer_path': self.pointer.path
        }

try:
    # For Python 3.0 and later
    from urllib.request import urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen

log = logging.getLogger(__name__)

def set_backend(dag, backend, proxymaker = None):
    '''
    sets backend and proxies for each node in the DAG.
    proxymaker is a 1-ary function that takes the node object and
    returns a suitable result proxy

    :param dag: the dag object (par of the workflow object)
    :param backend: the backend to associate with the nodes
    :param proxymaker: function object with signature f(node) to create result proxies
    '''
    for nodename in dag.nodes():
        n = dag.getNode(nodename)
        n.backend = backend
        n.resultproxy = proxymaker(n) if proxymaker else None
        n.update_state()

DEFAULT_ID_METHOD = 'uuid'


def json_hash(jsonable):
    '''
    returns a content-based hash of a json-able structure (serialized with WithJsonRefEncoder)

    :param jsonable: a json-serializable object
    :return: the hash
    '''
    the_hash  = hashlib.sha1(json.dumps(jsonable, cls=WithJsonRefEncoder, sort_keys=True).encode('utf-8')).hexdigest()
    return the_hash


def get_id_fromjson(jsonobject, method = DEFAULT_ID_METHOD):
    method = os.environ.get('YADAGE_ID_METHOD', method)
    if method == 'uuid':
        return str(uuid.uuid4())
    elif method == 'jsonhash':
        return json_hash(jsonobject)
    else:
        raise NotImplementedError(
            'unkown id generation method {}'.format(method))

def get_obj_id(obj_with_json_method, method = DEFAULT_ID_METHOD):
    return get_id_fromjson(obj_with_json_method.json(), method)

def process_refs(x, dag):
    if type(x) == dict:
        for k, v in x.items():
            x[k] = process_refs(v, dag)
        return x
    elif type(x) == list:
        for i, e in enumerate(x):
            x[i] = process_refs(e, dag)
        return x
    elif type(x) == outputReference:
        return x.pointer.resolve(dag.getNode(x.stepid).result)
    else:
        return x

def leaf_iterator(jsonable):
    '''
    generator function to yield leafs items of a JSON-like structure alongside
    their position in the structure as determined by a JSONPointer.

    :param jsonable: a json-serializable object
    :return: tuples (jsonpointer, leaf value)
    '''
    allleafs = jq.jq('leaf_paths').transform(jsonable, multiple_output = True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x) for x in allleafs]
    for x in leafpointers:
        yield x,x.get(jsonable)

def leaf_iterator_jsonlike(jsonlike, path = None):
    ''''
    :param jsonlike: a jsonlike object (i.e. nested lists/arrays: leaf-types must not be JSONable)
    :return: iterator that yields (JSON path, leaf value) tuples
    '''
    path = path or []
    if type(jsonlike) == list:
        for i,x in enumerate(jsonlike):
            thispath = copy.deepcopy(path)
            thispath.append(i)
            for leaf in leaf_iterator_jsonlike(x, path = thispath):
                yield leaf
    elif type(jsonlike) == dict:
        for k,v in jsonlike.items():
            thispath = copy.deepcopy(path)
            thispath.append(k)
            for leaf in leaf_iterator_jsonlike(v, path = thispath):
                yield leaf
    else:
        yield jsonpointer.JsonPointer.from_parts(path),jsonlike


def process_jsonlike(jsonlike, jq_obj_selector, callback):
    wflowrefs = [jsonpointer.JsonPointer.from_parts(x) for x in jq.jq(
            'paths(if objects then {} else false end)'.format(jq_obj_selector)
            ).transform( jsonlike, multiple_output = True
    )]
    for wflowref in wflowrefs:
        value = callback(wflowref.resolve(jsonlike))
        wflowref.set(jsonlike,value)
    return jsonlike

def pointerize(typedleafs, asref=False, stepid=None):
    '''
    a helper method that replaces leaf nodes in a JSON object with
    a outputReference objects (~ a JSONPath) pointing to that leaf position
    useful to track access to leaf nodes later on.
    '''
    pointerized = typedleafs.copy().json()
    for p,v in typedleafs.leafs():
        newval = outputReference(stepid, p) if asref else {'$wflowpointer': {'step': stepid,'result': p.path}} if stepid else p.path
        p.set(pointerized, newval)
    return pointerized

def options_from_eqdelimstring(opts):
    options = {}
    for x in opts:
        key, value = x.split('=')
        options[key] = yaml.load(value)
    return options

def getinit_data(initfiles, parameters):
    '''
    get initial data from both a list of files and a list of 'pname=pvalue'
    strings as they are passed in the command line <pvalue> is assumed to be a
    YAML parsable string.
    '''

    initdata = {}
    for initfile in initfiles:
        log.info('loading initialization data from file %s',initfile)
        initdata.update(**yaml.load(open(initfile)))

    initdata.update(**options_from_eqdelimstring(parameters))
    return initdata

def prepare_workdir_from_archive(initdir, inputarchive):
    if os.path.exists(initdir):
        raise click.exceptions.ClickException(click.style(
            "initialization directory exists and input archive give. Can't have both", fg='red'))
    os.makedirs(initdir)
    localzipfile = '{}/.yadage_inputarchive.zip'.format(initdir)
    f = urlopen(inputarchive)
    with open(localzipfile,'wb') as lf:
        lf.write(f.read())
    with zipfile.ZipFile(localzipfile) as zf:
        zf.extractall(path=initdir)
    os.remove(localzipfile)

def setupbackend_fromstring(backend, backendopts = None):
    backendopts = backendopts or {}
    log.info('setting up backend %s with opts %s', backend, backendopts)
    import yadage.backends.packtivitybackend as pb
    return pb.PacktivityBackend(
            packtivity_backendstring = backend,
            backendopts = backendopts
    )

def state_provider_from_string(dataarg,dataopts = None):
    dataopts = dataopts or {}
    log.info('%s %s',dataarg,dataopts)
    if len(dataarg.split(':',1)) == 1:
        dataarg = 'local:'+dataarg
    if dataarg.startswith('local:'):
        import yadage.state_providers.localposix
        return yadage.state_providers.localposix.setup_provider(dataarg,dataopts)
    if dataarg.startswith('py:'):
        _,module, setupfunc,dataarg = dataarg.split(':',3)
        module = importlib.import_module(module)
        setupfunc = getattr(module,setupfunc)
        return setupfunc(dataarg,dataopts)
    if dataarg.startswith('fromenv:'):
        module = importlib.import_module(os.environ['PACKTIVITY_STATEPROVIDER'])
        return module.setup_provider(dataarg,dataopts)
    raise RuntimeError('unknown data type %s %s', dataarg, dataopts)

def get_init_spec(discover):
    return {
        'process': None,
        'environment': None,
        'publisher': {
            'publisher_type': 'fromparjq-pub',
            'script': '.',
            'tryExact': True,
            'glob': discover,
            'relative_paths': True
        }
    }

def init_stage_spec(parameters, discover, used_inputs, name, nodename = None):
    return {
        'name': name,
        'dependencies': {
            "dependency_type": "jsonpath_ready",
            "expressions": []
        },
        'scheduler': {
             'scheduler_type': 'init-stage',
             'parameters': parameters,
             'inputs':   used_inputs,
             'nodename': nodename,
             'discover': discover
        }
    }

def rule_steps_indices(workflow):
    rule_to_steps_index = {}
    steps_to_rule_index = {}
    for rule in workflow.rules + workflow.applied_rules:
        path = '/'.join([rule.offset, rule.rule.name])
        p = jsonpointer.JsonPointer(path)
        try:
            steps_of_rule = [x['_nodeid'] for x in p.resolve(workflow.stepsbystage) if '_nodeid' in x]
        except jsonpointer.JsonPointerException:
            steps_of_rule = []
        rule_to_steps_index[rule.identifier] = steps_of_rule
        for step in steps_of_rule:
            steps_to_rule_index[step] = rule.identifier
    return rule_to_steps_index, steps_to_rule_index

def advance_coroutine(coroutine):
    try:
        return coroutine.next()
    except AttributeError:
        return coroutine.__next__()



import json
import logging
import os
import time

import yadage.backends.federatedbackend as federatedbackend
from packtivity.statecontexts import load_state
from packtivity.typedleafs import TypedLeafs
from yadage.utils import json_hash

from ..backends import CachedProxy
from .trivialbackend import TrivialBackend, TrivialProxy

log = logging.getLogger(__name__)

def setupcache_fromstring(configstring):
    '''
    generate cache from a string configuration (such as those passed from CLIs).
    valid strings are:

    - checksums:<path to cache file (JSON)>

    :param configstring: the configuration string
    :return: a cache object
    '''

    configparts  = configstring.split(':',1)
    strategy, specificconf = configparts
    if strategy == 'checksums':
        configfile = specificconf
        log.info('checksums caching strategy')
        return ChecksumCache(configfile)
    else:
        raise RuntimeError('unknown caching config')

class CachedBackend(federatedbackend.FederatedBackend):
    '''
    A caching backend that takes an existing backend and a cache configuration.
    If at submission time a cache for a given task is valid, the result will be
    returned directly via a trivial proxy. If not submission proceeds as normal
    '''

    def __init__(self, backend, cache):
        super(CachedBackend, self).__init__({
            'cache': TrivialBackend(),
            'primary': backend
        })
        self.cache = cache
        self.primary_enabled = True

    def ready(self, proxy):
        isready = super(CachedBackend, self).ready(proxy)
        #if this a genuinely new result, cache it under the task cache id
        if isready and type(proxy) is not TrivialProxy:
            result = super(CachedBackend, self).result(proxy)
            status = super(CachedBackend, self).successful(proxy)
            if not self.cache.cacheexists(proxy.cacheid):
                self.cache.cacheresult(proxy.cacheid, status, result)
        return isready

    def expected_result(self, task):
        return self.backends['packtivity'].prepublish(task)

    def routedsubmit(self, task):
        cached = self.cache.cacheddata(task)
        if cached:
            log.info('use cached result for task: %s',task.metadata['name'])
            return TrivialProxy(cached['status'], cached['result'].json(), task.state.datamodel)
        else:
            if not self.primary_enabled:
                raise RuntimeError('cache failed but refusing to submit to primary since it was explicitly disabled')
            log.info('do proper submit for task: %s',task.metadata['name'])
            #create id for this task using the cache builder, with which
            #we will store the result with once it's ready
            cacheid = self.cache.cacheid(task)

            task.state.reset()
            #maybe we erroneously are re-submitting this see
            #https://github.com/diana-hep/yadage/issues/45 TODO!
            primaryproxy = self.backends['primary'].submit(task.spec, task.parameters, task.state, task.metadata)
            cachedproxy  = CachedProxy(primaryproxy, cacheid)
            return cachedproxy

    def routeproxy(self, proxy):
        if type(proxy) == TrivialProxy:
            return 'cache', proxy
        else:
            return 'primary', proxy.proxy


class CacheBuilder(object):
    '''
    Cache base class. Implementations need to provide cachevalid and
    generate_validation_data methods
    '''
    def __init__(self, cachefile):
        self.cachefile = cachefile
        if not os.path.exists(self.cachefile):
            log.info('fresh cache as file at %s does not exist yet.',self.cachefile)
            self.cache = {}
        else:
            log.info('reading cache from %s',self.cachefile)
            self.cache = json.load(open(self.cachefile))

    def todisk(self):
        log.info('writing cache to %s',self.cachefile)
        json.dump(self.cache, open(self.cachefile, 'w'), indent=4, sort_keys=True)

    def remove(self,cacheid):
        self.cache.pop(cacheid)

    def cacheresult(self, cacheid, status, result):
        '''
        saves a result and its status under a unique identifier and includes
        validation data to verify the cache validity at a later point in time
        '''
        log.debug('caching result for cacheid: %s',cacheid)
        log.debug('caching result for process: %s',self.cache[cacheid]['task']['spec']['process'])
        self.cache[cacheid]['result'] = {
            'status': 'SUCCESS' if status else 'FAILED',
            'result': result.json(),
            'cachingtime': time.time(),
            'validation_data': self.generate_validation_data(cacheid)
        }
        self.todisk()

    def cachedresult(self,cacheid, state, silent = True):
        '''
        returns the cached result. when silent = True the mthod exits gracefully
        and returns None
        '''
        if silent:
            if not self.cacheexists(cacheid): return None
        return {
            'result': TypedLeafs(self.cache[cacheid]['result']['result'], state.datamodel),
            'status': self.cache[cacheid]['result']['status']
        }

    def cacheid(self,task):
        t = task.json()
        tocache = [t['state'], t['parameters'], t['spec']]
        hash = json_hash(tocache)
        # log.info('%s -> %s', json.dumps(tocache), hash)
        return hash

    def cacheexists(self,cacheid):
        return cacheid in self.cache and 'result' in self.cache[cacheid]

    def cacheddata(self, task, remove_invalid = True):
        '''
        returns results from a valid cache entry if it exists, else None
        if remove_invalid = True, also removes invalid cache entries from cache
        '''
        cacheid = self.cacheid(task)
        #register this task with the cacheid if we don't know about it yet
        log.debug('checking cache for task %s',task.metadata['name'], cacheid)
        if cacheid not in self.cache:
            self.cache[cacheid] = {'task' : task.json()}
        if not self.cacheexists(cacheid):
            log.info('cache non-existent for task %s (%s)',task.metadata['name'],cacheid)
            return None
        if not self.cachevalid(cacheid):
            self.remove(cacheid)
            self.cache[cacheid] = {'task' : task.json()}
            return None
        #return a cached result if we have one if not, return None
        result =  self.cachedresult(cacheid, task.state, silent = True)
        log.debug('returning cached result %s',result)
        return result

    def cachevalid(self, cacheid):
        raise NotImplementedError

    def generate_validation_data(self,cacheid):
        raise NotImplementedError

class ChecksumCache(CacheBuilder):
    '''
    This is a cache that based on state hashes
    '''

    def __init__(self, cachefile):
        super(ChecksumCache, self).__init__(cachefile)

    def remove(self,cacheid):
        '''
        removes entry from the cache and reset its state

        :param cachid: the cache entry identifier
        '''
        task = self.cache[cacheid]['task']
        log.debug('removing cache entry %s and resetting state',cacheid)
        load_state(task['state']).reset()
        super(ChecksumCache, self).remove(cacheid)


    def generate_validation_data(self,cacheid):
        '''
        calculate validation data (checksums in SHA1 hash form)

        :param cachid: the cache entry identifier
        '''
        validation_data = {
            'state_hash': load_state(self.cache[cacheid]['task']['state']).state_hash()
        }

        log.debug('validation data is %s',validation_data)
        return validation_data

    def cachevalid(self, cacheid):
        '''
        validate cache entry

        :param cachid: the cache entry identifier
        '''
        # task = self.cache[cacheid]['task']

        stored_validation_data = self.cache[cacheid]['result']['validation_data']
        validation_data_now = self.generate_validation_data(cacheid)

        if not stored_validation_data == validation_data_now:
            log.info('cache invalid')
            return False
        log.info('cache valid')
        return True



import logging
import shlex
import subprocess

import jsonpointer
import pydotplus

log = logging.getLogger(__name__)


def fillscope(cluster, workflow, scope='', subcluster=True):
    # scopecluster = stagecluster = pydotplus.graphviz.Cluster(graph_name =
    # '_'.join(stagescopeprts),

    if subcluster:
        scopecluster = pydotplus.graphviz.Cluster(
            graph_name=scope.replace('/', ''),
            label=''.join(['[{}]'.format(p) for p in scope.split('/')[1:]]),
            style='solid',
            color='blue')
        cluster.add_subgraph(scopecluster)
    else:
        scopecluster = cluster
    scopeptr = jsonpointer.JsonPointer(scope)
    scoped = scopeptr.resolve(workflow.stepsbystage)
    for stage, elements in scoped.items():
        stagescopeprts = scopeptr.parts + [stage]

        if subcluster:
            stagecluster = pydotplus.graphviz.Cluster(
                graph_name='_'.join(stagescopeprts),
                label=stage,
                labeljust='l',
                color='grey',
                style='dashed')
            scopecluster.add_subgraph(stagecluster)
        else:
            stagecluster = scopecluster

        for i, element in enumerate(elements):
            if '_nodeid' in element:
                element = element['_nodeid']
                targetcl = stagecluster if stage != 'init' else scopecluster
                shape = 'diamond' if stage in ['init','output'] else 'box'
                label = '' if stage in ['init','output'] else '{}[{}]'.format(stage, i)
                additional = {'fixedsize': True, 'height': 0.2,
                              'width': 0.2} if stage in ['init','output'] else {}
                targetcl.add_node(pydotplus.graphviz.Node(
                    element, label=label, color='blue', shape=shape, **additional))
                add_result(targetcl, element,
                           workflow.dag.getNode(element).result)
            elif type(element) == dict:
                # recurse...
                fillscope(
                    stagecluster, workflow, jsonpointer.JsonPointer.from_parts(
                        scopeptr.parts + [stage, i]).path,
                    subcluster=subcluster
                )


def path_to_id(stepid, path):
    return '{}_{}'.format(stepid, path.replace('/', '_'))

def add_result(graph, parent, typedleafs):
    leafpointers = [p for p,v in typedleafs.leafs()]

    for leaf in leafpointers:
        leafid = path_to_id(parent, leaf.path)
        # value = leaf.resolve(jsondata)
        source = ''.join(['[{}]'.format(p) for p in leaf.parts])
        label = '{}'.format(source)
        graph.add_node(pydotplus.graphviz.Node(
            leafid, label=label, color='red'))
        graph.add_edge(pydotplus.graphviz.Edge(parent, leafid))


def attach_to_results(provgraph, workflow, node):
    for dep in workflow.dag.getNode(node).task.inputs:
        resultid = path_to_id(dep.stepid, dep.pointer.path)
        provgraph.add_edge(pydotplus.graphviz.Edge(resultid, node))


def connect(provgraph, workflow):
    for node in workflow.dag.nodes():
        attach_to_results(provgraph, workflow, node)


def provdotgraph(workflow, subcluster=True):
    provgraph = pydotplus.graphviz.Graph()
    fillscope(provgraph, workflow, subcluster=subcluster)
    connect(provgraph, workflow)
    return provgraph


def write_prov_graph(workdir, workflow, vizformat='pdf'):
    provgraph = provdotgraph(workflow, subcluster=True)

    with open('{}/yadage_workflow_instance.dot'.format(workdir), 'w') as dotfile:
        dotfile.write(provgraph.to_string())

    with open('{}/yadage_workflow_instance.{}'.format(workdir, vizformat), 'w') as outfile:
        subprocess.call(shlex.split(
            'dot -T{} {}/yadage_workflow_instance.dot'.format(vizformat, workdir)), stdout=outfile)



import logging

import networkx as nx

import yadage.utils as utils

log = logging.getLogger(__name__)

def reset_node_state(node):
    node.submit_time = None
    node.ready_by_time = None
    node.resultproxy = None
    node.backend = None

def reset_step(workflow, step):
    log.debug('resetting %s', step)
    s = workflow.dag.getNode(step)
    reset_node_state(s)
    try:
        s.task.state.reset()
    except AttributeError:
        pass

def remove_rule(workflow, ruleid):
    assert ruleid in [x.identifier for x in workflow.rules]
    r = workflow.view().getRule(identifier = ruleid)
    workflow.rules.remove(r)
    workflow.view(r.offset).bookkeeper['_meta']['stages'].remove(ruleid)

def remove_rules(workflow, ruleids):
    for r in ruleids:
        remove_rule(workflow, r)

def undo_rule(workflow, ruleid):
    r2s, s2r   = utils.rule_steps_indices(workflow)

    if not ruleid in [r.identifier for r in workflow.applied_rules]:
        log.info('rule not in list of applied rules. possibly already undone during recursion.')
        return

    down_rules = list(set([s2r[s] for s in collective_downstream(workflow, r2s[ruleid])]))

    assert ruleid not in down_rules
    #TODO: not an issue now, but if stages schedule linked nodes we might
    #we might see downstream steps that are part of the same rule

    if not down_rules:
        r = workflow.view().getRule(identifier = ruleid)
        log.info('undoing %s', r)
        idx   = workflow.applied_rules.index(r)
        stepids = r2s[ruleid]
        steps = [workflow.dag.getNode(nid) for nid in stepids]

        # reset states of the step
        reset_steps(workflow,stepids)
        #
        # delete the steps
        for n in steps:
            workflow.dag.removeNode(n)

        #remove rule from applied list
        r = workflow.applied_rules.pop(idx)

        #reset all index data for stage
        workflow.view(r.offset).steps.pop(r.rule.name)

        #remove steps and stage from bookkeeping
        for s in stepids:
            workflow.view(r.offset).bookkeeper['_meta']['steps'].remove(s)

        workflow.view(r.offset).bookkeeper['_meta']['stages'].remove(ruleid)

        #re-append the rule
        workflow.view(r.offset).addRule(r.rule)
        return

    while True:
        if not down_rules:
            break
        for r in down_rules:
            undo_rule(workflow,r)
        r2s, s2r   = utils.rule_steps_indices(workflow)
        down_rules = list(set([s2r[s] for s in collective_downstream(workflow, r2s[ruleid])]))
    undo_rule(workflow,ruleid)

def undo_rules(workflow, ruleids):
    log.debug('undo rules')
    for r in ruleids:
        undo_rule(workflow, r)

def reset_steps(workflow, steps):
    for s in steps:
        reset_step(workflow, s)

def collective_downstream(workflow, steps):
    downstream = set()
    for step in steps:
        for x in nx.descendants(workflow.dag, step):
            downstream.add(x)
    return list(downstream)



import datetime
import os
import time

import jsonpointer

import adage.node
import packtivity
import yadage.tasks as tasks


class YadageNode(adage.node.Node):
    '''
    Node object for yadage that extends the default with
    the ability to have prepublished results
    '''

    def __init__(self, name, task, identifier=None):
        super(YadageNode, self).__init__(name, task, identifier)

    def __repr__(self):
        lifetime = datetime.timedelta(seconds = (time.time() - self.define_time))
        runtime = None
        if self.state != adage.nodestate.DEFINED:
            referencetime = time.time() if not self.ready() else self.ready_by_time
            runtime = datetime.timedelta(seconds = (referencetime - self.submit_time))
        return '<YadageNode {} {} lifetime: {}  runtime: {} (id: {}) has result: {}>'.format(
            self.name, self.state, lifetime, runtime, self.identifier, self.has_result()
        )

    def has_result(self):
        if 'YADAGE_IGNORE_PREPUBLISHING' in os.environ:
            return self.successful()
        return (self.expected_result is not None) or self.successful()

    @property
    def expected_result(self):
        if 'YADAGE_IGNORE_PREPUBLISHING' in os.environ:
            return
        try:
            return self._prepublished_cache
        except AttributeError:
            self._prepublished_cache = packtivity.prepublish_default(self.task.spec, self.task.parameters.json(), self.task.state)
            return self._prepublished_cache

    @property
    def result(self):
        if self.expected_result is not None and 'YADAGE_IGNORE_PREPUBLISHING' not in os.environ:
            if self.ready() and self.successful():
                sanity =  super(YadageNode, self).result.json() == self.expected_result.json()
                if not sanity:
                    raise RuntimeError('prepublished and actual result differ:\n result:\n{}\n prepub:{}'.format(
                        super(YadageNode, self).result.json(),self.expected_result.json())
                )
            return self.expected_result
        return super(YadageNode, self).result

    def readfromresult(self,pointerpath, whoisreading = None, failsilently = False):
        if not self.has_result():
            if failsilently: return None
            raise RuntimeError('attempt')
        pointer = jsonpointer.JsonPointer(pointerpath)
        if whoisreading:
            whoisreading.inputs.append(tasks.outputReference(self.identifier,pointer))
        v = pointer.resolve(self.result)
        return v

    @classmethod
    def fromJSON(cls, data, state_deserializer):
        if data['task']['type'] == 'packtivity_task':
            task = tasks.packtivity_task.fromJSON(data['task'], state_deserializer)
            return cls(data['name'], task, data['id'])
        else:
            raise RuntimeError('unknown task type',data['task']['type'])



import logging

from .handlers.predicate_handlers import handlers as pred_handlers
from .utils import get_id_fromjson

log = logging.getLogger(__name__)


class OffsetStage(object):
    '''
    A wrapper object around a scoped rule, so that it can be applied from
    a global p.o.v., i.e. as adage expects its rules.
    '''

    def __init__(self, rule, offset=None, identifier=None):
        '''
        initializes a scoped rule. scope is defined by a JSONPointer, e.g.
        one[0]two.three

        '''

        self.rule = rule
        self.offset = offset
        self.identifier = identifier or get_id_fromjson({
            'rule': rule.json(),
            'offset': offset
        })

    def __repr__(self):
        return '<OffsetStage {}/{} >'.format(self.offset,self.rule.name)

    def applicable(self, adageobj):
        '''
        determin whether the rule is applicable. Evaluated within the offset.
        :param adageobj: the workflow object
        '''
        from .wflowview import WorkflowView #importing here to avoid circdep
        x = self.rule.applicable(WorkflowView(adageobj, self.offset))
        return x

    def apply(self, adageobj):
        '''
        applies a rule within the scope set by offset

        :param adageobj: the workflow object
        '''
        from .wflowview import WorkflowView #importing here to avoid circdep
        self.rule.apply(WorkflowView(adageobj, self.offset))

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data, state_provider_deserializer = None):
        if data['rule']['type'] == 'JsonStage':
            rule = JsonStage.fromJSON(data['rule'], state_provider_deserializer)
        else:
            RuntimeError('unknown stage type %s', data['rule']['type'])
        return cls(
            rule=rule,
            identifier=data['id'],
            offset=data['offset']
        )

    def json(self):
        return {'type': 'offset',
                'id': self.identifier,
                'offset': self.offset,
                'rule': self.rule.json()}


class ViewStageBase(object):
    '''
    base class for workflow stages operating on workflow views, that may
    expose only a partial slice of an overall workflow.
    The class also provides common methods to ease modifying the workflow.

    Implementations are required to provide
    * a ready() method to implement the predicate method
    * a schedule() method that is called upon apply
    '''
    def __init__(self, name, state_provider):
        self.name = name
        self.state_provider = state_provider
        self.view = None

    def schedule(self):
        raise NotImplementedError()

    def ready(self):
        raise NotImplementedError()

    def applicable(self, flowview):
        self.view = flowview
        return self.ready()

    def apply(self, flowview):
        self.view = flowview
        self.schedule()

    def addStep(self, step):
        dependencies = [self.view.dag.getNode(k.stepid) for k in step.inputs]
        return self.view.addStep(step, stage = self.name, depends_on=dependencies)

    def addWorkflow(self, rules, isolate = True):
        self.view.addWorkflow(rules, stage=self.name if isolate else None)

    def json(self):
        return {
            'name': self.name,
            'state_provider': self.state_provider.json() if self.state_provider else None
        }

class JsonStage(ViewStageBase):
    '''
    A stage that is defined via the JSON scheduler schemas
    '''

    def __init__(self, json, state_provider):
        self.stagespec = json['scheduler']
        self.depspec = json['dependencies']
        super(JsonStage, self).__init__(json['name'], state_provider)

    def __repr__(self):
        return '<JsonStage: {}>'.format(self.name)

    def ready(self):
        if not self.depspec:
            return True
        predicate = pred_handlers[self.depspec['dependency_type']]
        return predicate(self, self.depspec, self.stagespec)

    def schedule(self):
        #imported here to avoid circular dependency
        from .handlers.scheduler_handlers import handlers as sched_handlers
        scheduler = sched_handlers[self.stagespec['scheduler_type']]
        scheduler(self, self.stagespec)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data, state_provider_deserializer):
        return cls(
            json={
                'scheduler': data['scheduler'],
                'name': data['name'],
                'dependencies': data['dependencies']
            },
            state_provider=state_provider_deserializer(data['state_provider'])
        )

    def json(self):
        data = super(JsonStage, self).json()
        data.update(type='JsonStage', scheduler=self.stagespec, dependencies = self.depspec)
        return data



import yadageschemas


def workflow(source, toplevel, schema_name='yadage/workflow-schema', schemadir=None, validate=True):
    '''
    load (and validate) workflow from source
    param source: URI fragment to the source
    param toplevel: base URI to resolve references from
    schemadir
    '''
    dialect, spec = ('raw_with_defaults', source) if ':' not in source else source.split(':',1)

    specopts = {
        'toplevel': toplevel,
        'schema_name': schema_name,
        'schemadir': schemadir,
        'load_as_ref': False
    }
    validopts = {
        'schema_name': schema_name,
        'schemadir': schemadir,
    }

    data = yadageschemas.load(
        spec, specopts, validate = True, validopts = validopts, dialect = dialect
    )
    return data

def validate(data, schema_name='yadage/workflow-schema', schemadir=None):
    '''validate workflow data'''
    validopts = {
        'schema_name': schema_name,
        'schemadir': schemadir,
    }
    yadageschemas.validate_spec(data,validopts)



class FederatedBackend(object):
    '''
    A meta backend that routes tasks to several internal backends
    '''

    def __init__(self, backends):
        '''takes a dictionary of backendname->backendobject and a router'''
        self.backends = backends

    def routedsubmit(self, task):
        raise NotImplementedError('needs implementation')

    def routeproxy(self, proxy):
        raise NotImplementedError('needs implementation')

    def submit(self, task):
        return self.routedsubmit(task)

    def result(self, proxy):
        b,p = self.routeproxy(proxy)
        return self.backends[b].result(p)

    def expected_result(self, proxy):
        b,p = self.routeproxy(proxy)
        return self.backends[b].expected_result(p)

    def ready(self, proxy):
        b,p = self.routeproxy(proxy)
        return self.backends[b].ready(p)

    def successful(self, proxy):
        b,p = self.routeproxy(proxy)
        return self.backends[b].successful(p)

    def fail_info(self, proxy):
        b,p = self.routeproxy(proxy)
        return self.backends[b].fail_info(p)



#!/usr/bin/env python
import functools
import logging
import os

import click
import yaml

import yadage.manualutils as manualutils
import yadage.utils as utils
import yadage.workflow_loader as workflow_loader

from .steering_api import execute_steering
from .steering_object import YadageSteering

log = logging.getLogger(__name__)
logging.basicConfig(level = logging.INFO)


def connection_options(func):
    @click.option('-m', '--metadir', default='yadagemeta', help = 'directory to store workflow metadata')
    @click.option('--accept-metadir/--no-accept-metadir', default=True)
    @click.option('-r', '--controller', default='frommodel')
    @click.option('-o', '--ctrlopt', multiple=True, default=None, help = 'options for the workflow controller')
    @click.option('-s', '--modelsetup', default='filebacked:yadage_state.json')
    @click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
    @click.option('-b', '--backend', default='celery')
    @click.option('--local/--remote', default = True)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

def common_options(func):
    @click.option('-v', '--verbosity', default='ERROR')
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

@click.group()
def mancli():
    pass

@mancli.command()
@click.option('-s', '--modelsetup', default='filebacked:yadage_state.json')
@click.option('-t', '--toplevel', default=os.getcwd())
@click.option('-d', '--dataopt', multiple=True, default=None, help = 'options for the workflow data state')
@click.option('--metadir', default=None, help = 'directory to store workflow metadata')
@click.option('--parameter', '-p', multiple=True)
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initfiles', nargs = -1)
def init(workdir, workflow, initfiles, modelsetup, dataopt, metadir, toplevel, parameter):
    initdata = utils.getinit_data(initfiles, parameter)
    dataopts = utils.options_from_eqdelimstring(dataopt)

    ys = YadageSteering.create(
        dataarg = workdir,
        dataopts = dataopts,
        workflow = workflow,
        toplevel = toplevel,
        initdata = initdata,
        metadir = metadir,
        modelsetup = modelsetup,
    )
    assert ys

def click_print_applicable_stages(controller):
    click.secho('Applicable Stages: ', fg='blue')
    for x in controller.adageobj.rules:
        if x.identifier in controller.applicable_rules():
            click.secho('{}/{}'.format(x.offset, x.rule.name))

def click_print_submittable_nodes(controller):
    click.secho('Submittable Nodes: ', fg='blue')
    _, s2r = utils.rule_steps_indices(controller.adageobj)
    for x in controller.adageobj.dag.nodes():
        node = controller.adageobj.dag.getNode(x)
        rule = controller.adageobj.view().getRule(identifier = s2r[node.identifier])
        if node.identifier in controller.submittable_nodes():
            click.secho('node: {} ({}) part of stage {}'.format(node.name, node.identifier,  '/'.join([rule.offset,rule.rule.name])))

def click_print_rule(rule, step_index, step_status):
    steps_of_rule = step_index[rule.identifier]

    rule_stats = {}
    for x in steps_of_rule:
        rule_stats.setdefault(step_status[x],0)
        rule_stats[step_status[x]] += 1
    click.secho('{}/{} ({})' .format(rule.offset, rule.rule.name, '/'.join('{}: {}'.format(k,v) for k,v in rule_stats.items())))

def click_print_applied_stages(controller):
    click.secho('Applied Stages: ', fg='blue')
    r2s, s2r = utils.rule_steps_indices(controller.adageobj)
    step_status = {s: str(controller.adageobj.dag.getNode(s).state) for s in s2r.keys()}
    # print(step_status)
    for x in controller.adageobj.applied_rules:
        click_print_rule(x, r2s, step_status)

@mancli.command()
@click.argument('name', default=None, nargs = -1)
@connection_options
@common_options
def apply_stage(name,
          metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
          verbosity
         ):

    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    if not name:
        click_print_applicable_stages(controller)
        return

    for n in name:
        offset, scopedname = n.rsplit('/',1)
        rule = controller.adageobj.view(offset).getRule(scopedname)
        if not rule:
            click.secho('No such stage {}, pick one of the applicable below:'.format(n), fg='red')
            click_print_applicable_stages(controller)
            return

        if rule.identifier in [r.identifier for r in controller.adageobj.applied_rules]:
            click.secho('Stage {} was already applied.'.format(n), fg = 'yellow')
            continue

        if rule.identifier not in controller.applicable_rules():
            click.secho('Rule {} not yet applicable'.format(n), fg = 'red')
            continue

        controller.apply_rules([rule.identifier])
        click.secho('Stage {} applied'.format(n), fg = 'green')

@mancli.command()
@click.option('-n','--nodeid', default=None)
@click.option('-a','--allof', default=None)
@click.option('-o', '--offset', default='')
@connection_options
@common_options
def submit(nodeid, allof, offset,
           metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
           verbosity
           ):
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    if not (allof or nodeid):
        click_print_submittable_nodes(controller)
        return

    if nodeid:
        nodes_to_submit = [nodeid] if nodeid in controller.submittable_nodes() else []
    if allof:
        offset, name = allof.rsplit('/',1)
        rule = controller.adageobj.view().getRule(name = name, offset = offset)
        if not rule:
            click.secho('stage not found!', fg = 'red')
            return


        all_submittable = controller.submittable_nodes()
        _, s2r = utils.rule_steps_indices(controller.adageobj)
        nodes_to_submit = [x for x in all_submittable if s2r[x] == rule.identifier]

    if not nodes_to_submit:
        click.secho('No nodes to submit (perhaps already submitted?)', fg = 'yellow')
        return

    controller.submit_nodes(nodes_to_submit)
    click.secho('submitted: {}'.format(nodes_to_submit), fg = 'green')

@mancli.command()
@connection_options
@common_options
def shell(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
         verbosity
         ):
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    click.secho('launching yadage shell', fg = 'green')
    assert ys
    import IPython
    IPython.embed()

@mancli.command()
@connection_options
@common_options
@click.option('--show-processed/--hide-processed', default = False)
def show(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
         verbosity,
         show_processed
         ):

    ctrlarg = controller # for later
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    click.secho('''\
Workflow:
---------
state source: {statesource}
successful: {successful}
finished: {finished}
valid: {valid}
# of applicable rules: {applicable}
# of submittable nodes: {submittable}
'''.format(
        statesource = modelsetup if local else 'remote via {}'.format(ctrlarg),
        successful =  click.style(str(controller.successful()), fg = 'green' if controller.successful() else 'red'),
        finished = click.style(str(controller.finished()), fg = 'green' if controller.finished() else 'yellow'),
        valid = click.style(str(controller.validate()), fg = 'green' if controller.validate() else 'red'),
        applicable = len(controller.applicable_rules()),
        submittable = len(controller.submittable_nodes())
    ))
    click_print_applicable_stages(controller)
    click_print_submittable_nodes(controller)

    if show_processed:
        click_print_applied_stages(controller)



@mancli.command()
@click.argument('name')
@connection_options
@common_options
def preview(name,
            metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
            verbosity
            ):


    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    preview = manualutils.preview_rule(controller.adageobj, name)
    if not preview:
        click.secho('cannot preview {}'.format(name))
        raise click.Abort()

    new_rules, new_nodes = preview
    click.secho('Preview of Stage: # new rules: {} # new nodes {}'.format(
        len(new_rules), len(new_nodes)))
    for n in new_nodes:
        click.secho(
            '-> new node "{}" with {} upstream dependencies'.format(n['name'], len(n['parents'])))


def handle_common_options(verbosity):
    logging.basicConfig(level=getattr(logging, verbosity))

def handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local):
    ctrlopts = utils.options_from_eqdelimstring(ctrlopt)
    modelopts = utils.options_from_eqdelimstring(modelopt)
    ys = manualutils.connect(
        metadir = metadir,
        accept_metadir = accept_metadir,
        ctrlstring = controller,
        ctrlopts = ctrlopts,
        modelsetup = modelsetup,
        modelopts = modelopts,
        backendstring = backend if local else None
    )
    return ys


@mancli.command()
@click.option('--track/--no-track', default=True)
@click.option('-n', '--nsteps', default=-1, help = 'number of steps to process. use -1 to for no limit (will run workflow to completion)')
@click.option('-u', '--update-interval', default=1)
@click.option('-g','--strategy', help = 'set execution stragegy')
@connection_options
@common_options
def step(track , strategy, nsteps, update_interval,
         metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
         verbosity
         ):

    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)

    execute_steering(
        ys,
        updateinterval = update_interval,
        default_trackers = track,
        strategy = strategy
    )

@mancli.command()
@click.argument('dataarg')
@click.argument('workflow')
@click.option('-d', '--dataopts', multiple=True, default=None, help = 'ctrl opts for state provider')
@click.option('-f', '--offset', default = '')
@click.option('-g', '--groupname', default = None)
@click.option('-t', '--toplevel', default = os.curdir)
@click.option('--parameter', '-p', multiple=True)
@connection_options
@common_options
def add(offset, groupname, dataarg, dataopts, workflow, toplevel, parameter,
        metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
        verbosity
        ):
    parameter = utils.options_from_eqdelimstring(parameter)
    dataopts = utils.options_from_eqdelimstring(dataopts)
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    stages = []
    if parameter:
        inspec = utils.init_stage_spec(parameter, False, [], 'init')
        stages.append(inspec)

    stages = stages + workflow_loader.workflow(
        workflow,
        toplevel=toplevel,
        validate=True
    )['stages']
    controller.add_rules(stages, dataarg, offset = offset, groupname = groupname, dataopts = dataopts)

@mancli.command()
@click.option('-f', '--fileformat', default='pdf')
@click.option('-w', '--workdir', default=os.curdir)
@connection_options
@common_options
def visualize(workdir, fileformat,
              metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
              verbosity
            ):
    from .visualize import write_prov_graph

    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller
    if controller.backend:
        controller.sync_backend()


    write_prov_graph(workdir, controller.adageobj, fileformat)

@mancli.command()
@click.argument('name', default=None)
@click.option('-p','--patchspec', default=None)
@connection_options
@common_options
def edit_stage(name,patchspec,
          metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
          verbosity
    ):
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    if not name:
        click.secho('No stage specified. Pick one of the stages below:  ', fg = 'red')
        click_print_applied_stages(controller)
        return

    offset, scopedname = name.rsplit('/',1)
    rule = controller.adageobj.view(offset).getRule(scopedname)

    s = yaml.safe_dump(rule.rule.stagespec, default_flow_style = False)

    if patchspec:
        edited = yaml.load(open(patchspec))
    else:
        edited = yaml.load(click.edit(s, editor='vi'))
    controller.patch_rule(rule.identifier, edited)
    click.secho('updated {}'.format(name), fg = 'green')

@mancli.command()
@click.argument('name', default=None, nargs = -1)
@connection_options
@common_options
def undo_stage(name,
          metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
          verbosity
    ):
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    if not name:
        click.secho('No stage specified. Pick one of the stages below:  ', fg = 'red')
        click_print_applied_stages(controller)
        return

    for n in name:
        offset, scopedname = n.rsplit('/',1)
        rule = controller.adageobj.view(offset).getRule(scopedname)
        controller.undo_rules([rule.identifier])
        click.secho('undone {}'.format(n), fg = 'green')

@mancli.command()
@click.argument('name', default=None)
@connection_options
@common_options
def remove_stage(name,
          metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
          verbosity
    ):
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    if not name:
        click.secho('No stage specified. Pick one of the stages below:  ', fg = 'red')
        click_print_applied_stages(controller)
        return

    offset, name = name.rsplit('/',1)
    rule = controller.adageobj.view(offset).getRule(name)
    controller.remove_rules([rule.identifier])

@mancli.command()
@click.argument('name')
@connection_options
@common_options
def reset_stage(name,
          metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local,
          verbosity
    ):
    handle_common_options(verbosity)
    ys = handle_connection_options(metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local)
    controller = ys.controller

    offset, name = name.rsplit('/',1)
    rule = controller.adageobj.view(offset).getRule(name)
    if not rule:
        click.secho('state not found!', fg = 'red')

    r2s, _ = utils.rule_steps_indices(controller.adageobj)
    steps_of_rule = r2s[rule.identifier]
    controller.reset_nodes(steps_of_rule)

if __name__ == '__main__':
    mancli()



import logging

import yadage.backends.caching as caching
import yadage.backends.federatedbackend as federatedbackend
from packtivity.asyncbackends import ForegroundBackend
from packtivity.backendutils import backend_from_string

log = logging.getLogger(__name__)

class PacktivityBackend(federatedbackend.FederatedBackend):
    '''
    a backend that mainly submits step tasks to a packtivity backend
    except for init nodes, which are resolvable trivially.
    '''

    def __init__(self,  packtivity_backendstring = None, packtivity_backend = None, backendopts = None):
        if packtivity_backendstring:
            is_sync, backend = backend_from_string(packtivity_backendstring, backendopts = backendopts)
            assert not is_sync
        elif packtivity_backend:
            backend = packtivity_backend
        else:
            raise RuntimeError('need backend or backendstring')
        self.cached = False
        super(PacktivityBackend, self).__init__({
            'purepub': ForegroundBackend(),
            'packtivity': backend
        })

    def enable_cache(self,cachestring = None, cache = None):
        if not (cache or cachestring):
            raise RuntimeError('need to provide either cache object or cache config string')

        self.cached = True
        self.backends['packtivity'] = caching.CachedBackend(
            self.backends['packtivity'],
            cache = cache if cache else caching.setupcache_fromstring(cachestring)
        )

    def routedsubmit(self, task):
        is_pure_publishing = task.metadata['wflow_hints'].get('is_purepub',False)
        if is_pure_publishing:
            foreground_proxy = self.backends['purepub'].submit(
                task.spec, task.parameters, task.state, task.metadata
            )
            foreground_proxy.set_details({'labels': {'backend_hints': 'is_purepub'}})
            return foreground_proxy

        else:
            #this is a little hacky, because the packtivity backends
            #take unrolled spec/parameters/context while the adage API
            #takes generalized task objects
            #possibly could use Munch on the packtivity side to
            #dynammicaly create .task/.parameters/.state-able objects
            if self.cached:
                #Cached backends adhere to the task-based API
                return self.backends['packtivity'].submit(task)
            else:
                #primary packtivity backends adhere to the unrolled API
                return self.backends['packtivity'].submit(
                    task.spec, task.parameters, task.state, task.metadata
                )

    def expected_result(self, task):
        if self.cached:
            #Cached backends adhere to the task-based API
            return self.backends['packtivity'].expected_result(task)
        else:
            #primary packtivity backends adhere to the unrolled API
            return self.backends['packtivity'].prepublish(
                task.spec, task.parameters, task.state
            )

    def routeproxy(self, proxy):
        details = proxy.details() or {}
        if details.get('labels',{}).get('backend_hints') == 'is_purepub':
            return 'purepub', proxy
        else:
            return 'packtivity', proxy
        raise NotImplementedError('needs implementation')



import json
import logging
import shutil
import tempfile

import click
import yaml

from .backends.trivialbackend import TrivialBackend, TrivialProxy
from .handlers.expression_handlers import handlers as exh
from .utils import process_refs, set_backend
from .wflowstate import make_deserializer


def printRef(ref, dag, indent=''):
    click.secho('{}name: {} position: {}, value: {}, id: {}'.format(
        indent,
        dag.getNode(ref.stepid).name,
        ref.pointer.path,
        ref.pointer.resolve(dag.getNode(ref.stepid).result),
        ref.stepid
    ),
        fg='cyan')

def wflow_with_trivial_backend(instance,results):

    stateopts = {}
    wflowmaker = make_deserializer(stateopts)
    wflow = wflowmaker(json.load(open(instance)))

    resultdata = json.load(open(results))
    set_backend(
        wflow.dag,
        TrivialBackend(),
        proxymaker=lambda n: TrivialProxy(
            resultdata[n.identifier]['status'],
            resultdata[n.identifier]['result'],
            n.task.state.datamodel
        )
    )
    return wflow

@click.group()
def utilcli():
    pass


@utilcli.command()
@click.argument('instance')
@click.argument('results')
@click.argument('selection')
@click.option('-v', '--verbosity', default='INFO')
def testsel(instance, results, selection,verbosity):


    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


    wflow = wflow_with_trivial_backend(instance,results)

    selresult = exh['stage-output-selector'](wflow.view(), yaml.load(selection))

    if not selresult:
        click.secho('Bad selection', fg='red')
        return

    click.secho(json.dumps(
        process_refs(selresult, wflow.dag),
        sort_keys=True,
        indent=4,
        separators=(',', ': ')),
        fg='green'
    )


@utilcli.command()
@click.argument('instance')
@click.argument('results')
@click.argument('vizpdf')
def viz(instance, results, vizpdf):
    import yadage.visualize as visualize
    wflow = wflow_with_trivial_backend(instance,results)

    dirpath = tempfile.mkdtemp()
    visualize.write_prov_graph(dirpath, wflow)
    shutil.copy('{}/yadage_workflow_instance.pdf'.format(dirpath), vizpdf)
    shutil.rmtree(dirpath)


if __name__ == '__main__':
    utilcli()



import logging

import jsonpointer

import yadage.handlers.utils as utils

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()


def checkmeta(flowview, metainfo):
    log.debug('checking meta %s on view with offset %s',
              metainfo, flowview.offset)
    applied_ids = [rl.identifier for rl in flowview.applied_rules]
    rulesok = all([x in applied_ids for x in metainfo['stages']])

    stepsok = all([flowview.dag.getNode(x).has_result()
                   for x in metainfo['steps']])
    log.debug('all rules applied: %s, all steps have results: %s',
              rulesok, stepsok)
    return (rulesok and stepsok)


def scope_done(scope, flowview):
    '''
    walks recursively all scopes starting at some initial scope to determine if
    all steps and stages under this scope have been executed / applied. Will indicate
    that it's safe to reference any result of the workflow within that scope.
    '''
    log.debug('checking scope %s on view with offset %s',
              scope, flowview.offset)
    result = True

    bookkeeper = jsonpointer.JsonPointer(scope).resolve(flowview.bookkeeper)
    for k, v in bookkeeper.items():
        for k, v in bookkeeper.items():
            if k == '_meta':
                result = result and checkmeta(flowview, v)
            else:
                childscope = scope + '/{}'.format(k)
                result = result and scope_done(childscope, flowview)
    return result

@scheduler('jsonpath_ready')
def jsonpath_ready(stage, depspec,stagespec):
    '''
    the main predicate for yadage. for a list of jsonpath expressions
    determine whether the stage or workflow scope is ready (i.e. has a result)
    '''
    log.debug('checking jsonpath ready predicate\n%s', depspec)
    dependencies = depspec['expressions']
    for x in dependencies:
        depmatches = stage.view.query(x, stage.view.steps)
        if not depmatches:
            log.debug('no query matches, not ready')
            return False
        issubwork = '_nodeid' not in depmatches[0].value[0]
        if issubwork:
            log.debug('dependency is a subworkflow. determine if scope is done')
            if not all([scope_done(scope['_offset'], stage.view) for match in depmatches for scope in match.value]):
                return False
        else:
            if not all([x.has_result() for x in stage.view.getSteps(x)]):
                return False
    log.debug('all checks ok, predicate is True')
    return True

@scheduler('autodep')
def autodep(stage, depspec,stagespec):
    '''
    automatic dependency discovery based on stage spec
    '''
    raise NotImplementedError('automatic dependency not implemented yet')



import importlib
import logging
import os
from contextlib import contextmanager

import yadageschemas

from .steering_object import YadageSteering
from .strategies import get_strategy

log = logging.getLogger(__name__)

def run_workflow(*args, **kwargs):
    '''
    convenience function around steering context, when no additional settings
    are desired.
    '''
    with steering_ctx(*args, **kwargs):
        pass

def execute_steering(
    steering_object,
    updateinterval = 0.02,
    loginterval = 30,
    default_trackers=True,
    strategy = None,
    backend = None,
    cache = None
    ):

    ys = steering_object
    ys.adage_argument(
        default_trackers = default_trackers,
        trackevery = loginterval,
        update_interval = updateinterval,
    )

    if cache:
        if cache == 'checksums':
            backend.enable_cache(':'.join([cache,os.path.join(ys.metadir,'cache.json')]))
        else:
            backend.enable_cache(cache)


    custom_tracker = os.environ.get('YADAGE_CUSTOM_TRACKER',None)
    if custom_tracker:
        modulename,trackerclassname = custom_tracker.split(':')
        module = importlib.import_module(modulename)
        trackerclass = getattr(module,trackerclassname)
        ys.adage_argument(additional_trackers = [trackerclass()])

    if strategy is not None:
        ys.adage_argument(**get_strategy(strategy))

    ys.run_adage(backend)

@contextmanager
def steering_ctx(
    dataarg,
    workflow = None,
    initdata = None,
    toplevel = os.getcwd(),
    backend = None,
    controller = 'frommodel',
    ctrlopts = None,
    workflow_json = None,
    cache = None,
    dataopts = None,
    updateinterval = 0.02,
    loginterval = 30,
    schemadir = yadageschemas.schemadir,
    metadir = None,
    strategy=None,
    validate=True,
    visualize=True,
    accept_metadir = False,
    modelsetup = 'inmem',
    modelopts = None
):

    ys = YadageSteering.create(
        metadir = metadir, accept_metadir = True if (accept_metadir or cache) else False,
        dataarg = dataarg, dataopts = dataopts,
        workflow_json = workflow_json,
        workflow = workflow, toplevel = toplevel,
        schemadir = schemadir, validate = validate,
        initdata = initdata,
        modelsetup = modelsetup, modelopts = modelopts,
        controller = controller, ctrlopts = ctrlopts,
    )

    yield ys

    try:
        execute_steering(
            steering_object = ys,
            updateinterval = updateinterval,
            loginterval = loginterval,
            default_trackers = visualize,
            strategy = strategy,
            backend = backend,
            cache = cache
        )
    finally:
        ys.serialize()
    if visualize:
        ys.visualize()



import logging

from packtivity.typedleafs import TypedLeafs

from .utils import outputReference

log = logging.getLogger(__name__)

class packtivity_task(object):
    '''
    packtivity task
    '''
    def __init__(self, name, spec, state, parameters = None):
        self.metadata = {'name': name}
        self.inputs = []
        self.parameters = TypedLeafs(parameters or {}, state.datamodel if state else None)
        self.spec = spec
        self.state = state

    def used_input(self, reference):
        self.inputs += [reference]

    def used_inputs(self, references):
        for r in references: self.used_input(r)

    def pubOnlyTask(self):
        return (self.spec['environment'] is None) and (self.spec['process'] is None)

    def s(self, **parameters):
        self.parameters.update(**parameters)
        return self

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data, state_deserializer):
        instance = cls(
            data['metadata']['name'],
            data['spec'],
            state_deserializer(data['state']) if data['state'] else None,
            data['parameters']
        )
        instance.inputs       = map(outputReference.fromJSON, data['inputs'])
        instance.metadata.update(**data['metadata'])
        return instance

    def json(self):
        return {
            'metadata': self.metadata,
            'parameters': self.parameters.json(),
            'inputs': [x.json() for x in self.inputs],
            'type': 'packtivity_task',
            'spec': self.spec,
            'state': self.state.json() if self.state else None,
        }



import json
import logging

from packtivity.statecontexts import load_state

from .backends import load_proxy
from .state_providers import load_provider
from .wflow import YadageWorkflow
from .wflownode import YadageNode

log = logging.getLogger(__name__)


def make_deserializer(deserialization_opts = None):
    def deserializer(jsondata):
        workflow = YadageWorkflow.fromJSON(
            jsondata,
            lambda data: load_proxy(data,deserialization_opts),
            lambda data: load_provider(data,deserialization_opts),
            lambda data: YadageNode.fromJSON(data,
                                state_deserializer =
                                    lambda state: load_state(
                                        state,deserialization_opts
                                    )
            )
        )
        return workflow
    return deserializer

def load_model_fromstring(modelsetup,modelopts = None,initdata = None):
    modelopts = modelopts or {}
    if modelsetup.startswith('inmem'):
        return initdata
    elif modelsetup.startswith('filebacked'):
        filename = modelsetup.split(':')[-1]
        model   = FileBackedModel(
            filename = filename,
            initdata = initdata,
            deserializer = make_deserializer(modelopts)
        )
        return model
    elif modelsetup == 'mongo':
        model = MongoBackedModel(
            initdata = initdata,
            deserializer = make_deserializer(modelopts)
        )
        return model
    raise RuntimeError('unknown state model %s', modelsetup)

class MongoBackedModel(object):
    '''
    model that holds the workflow state in a MongoDB database.
    '''
    def __init__(self, deserializer = None, connect_string = 'mongodb://localhost:27017/', initdata = None, wflowid = None):
        from pymongo import MongoClient
        from bson.objectid import ObjectId
        self.deserializer = deserializer or make_deserializer()
        self.client = MongoClient(connect_string)
        self.db = self.client.wflowdb
        self.collection = self.db.workflows
        if initdata:
            insertion = self.collection.insert_one(initdata.json())
            self.wflowid = insertion.inserted_id
            log.info('created new workflow object with id %s', str(self.wflowid))
        if wflowid:
            self.wflowid = ObjectId(wflowid)

    def commit(self, data):
        '''
        :param data: data to commit to disk. needs to have '.json()' method
        return: None
        '''
        self.collection.replace_one({'_id':self.wflowid}, data.json())

    def load(self):
        '''
        param load: load data from database
        return: yadage workflow object
        '''
        return self.deserializer(self.collection.find_one({'_id':self.wflowid}))

class FileBackedModel(object):
    '''
    model that holds data on disk in a JSON file
    '''
    def __init__(self, filename, deserializer = None, initdata = None):
        self.filename = filename
        self.deserializer = deserializer or make_deserializer()
        if initdata:
            self.commit(initdata)

    def commit(self, data):
        '''
        :param data: data to commit to disk. needs to have '.json()' method
        '''
        log.debug('committing model')
        jsondata = data.json()

        with open(self.filename,'w') as statefile:
            json.dump(jsondata, statefile)

    def load(self):
        '''
        :return: the adage workflow object holding rules and the graph
        '''
        log.debug('loading model')
        with open(self.filename) as statefile:
            jsondata = json.load(statefile)
            return self.deserializer(jsondata)



import json


def snapshot(workflow, wflowfile, resultfile):
    '''
    snapshots the state of the workflow into two files:
        - a file holding the pure state of the workflow
        - a file holding the backend state, i.e. results to specific tasks
    '''
    with open(wflowfile, 'w') as wfile:
        json.dump(workflow.json(), wfile)




    with open(resultfile, 'w') as rfile:
        json.dump({
            k: {
                'result': workflow.dag.getNode(k).result.json(),
                'status': str(workflow.dag.getNode(k).state)
            } for k in workflow.dag.nodes()
        }, rfile
        )



import logging

import jsonpath_rw

import yadage.handlers.utils as utils

from ..utils import pointerize

log = logging.getLogger(__name__)

handlers, expression = utils.handler_decorator()



def select_reference(step, selection):
    '''
    resolves a jsonpath selection on a step's result data and returns JSONPointerized matches

    :param step: the step holding the result data
    :param selection: a JSONPath expression

    :return: the first (and single) match of the expression as a JSON pointer
    '''
    log.debug('selecting output from step %s', step)
    pointerized = pointerize(step['result'], asref=True, stepid=step['id'])
    matches = jsonpath_rw.parse(selection).find(pointerized)
    if not matches:
        log.error('no matches found for selection %s in result %s',
                  selection, step['result'])
        raise RuntimeError('no matches found in reference selection. selection %s | result %s', selection, step['result'])

    if len(matches) > 1:
        log.error('found multiple matches to query: %s within result: %s\n \ matches %s',
                  selection, step['result'], matches)
        raise RuntimeError('multiple matches in result jsonpath query')
    return matches[0].value

def combine_outputs(outputs, flatten, unwrapsingle):
    '''
    combines the result of multiple reference selections into a single outputs.
    non-list values will just be combined into a list while lists will be concatenated
    optinally we can return the sole element of a single-value list (argument: unwrapsingle)

    :param outputs:

    '''
    combined = []
    for reference in outputs:
        if type(reference) == list:
            if flatten:
                for elementref in reference:
                    combined += [elementref]
            else:
                combined += [reference]
        else:
            combined += [reference]
    if len(combined) == 1 and unwrapsingle:
        combined = combined[0]
    return combined


def select_steps(stageview, query):
    '''
    selects the step objects from the stage view based on a query and converts them to simple
    dictionaries with id,result keys

    :param stageview: the view object on which to perform the query
    :param query: a slection query (JSONPath expression)

    :return: list of {id: xx, result: yy} dictionaries
    '''
    return [{'id': s.identifier, 'result': s.result} for s in stageview.getSteps(query)]


def select_outputs(steps, selection, flatten, unwrapsingle):
    return combine_outputs(map(lambda s: select_reference(s, selection), steps), flatten, unwrapsingle)


@expression('stage-output-selector')
def stage_output_selector(stageview, selection):
    '''
    :param stageview: the workflow view objct
    :param selection: the JSON-like selection dictionary
    :return :
    resolves a output reference by selecting the stage and stage outputs
    '''
    log.debug('resolving selection %s', selection)
    if type(selection) is not dict:
        return None
    else:
        steps = select_steps(stageview, selection['stages'])
        log.debug('selected steps %s', steps)
        outputs = select_outputs(steps,
                                 selection['output'],
                                 selection.get('flatten', False),
                                 selection.get('unwrap', False))
        log.debug('selected outputs %s', outputs)
        return outputs

@expression('fromvalue')
def value_resolver(view, expression):
    value = view.getValue(expression['key'])
    log.debug('resolved to %s', value)
    if isinstance(value,dict) and 'expression_type' in value:
        log.debug('looking up expression %s', value)
        return handlers[value['expression_type']](view, value)
    else:
        log.info('not an expression value')
        return value



class CachedProxy(object):
    def __init__(self,proxy,cacheid):
        self.proxy = proxy
        self.cacheid = cacheid

    def details(self):
        return {}

    def json(self):
        return {'proxyname': 'CachedProxy', 'proxy': self.proxy.json(), 'cacheid': self.cacheid}

    @classmethod
    def fromJSON(cls,data,deserialization_opts = None):
        deserialization_opts = deserialization_opts or {}
        return cls(load_proxy(data['proxy'],deserialization_opts), data['cacheid'])

def load_proxy(data,deserialization_opts = None):
    deserialization_opts = deserialization_opts or {}
    import packtivity.backendutils
    import yadage.backends.packtivitybackend
    import yadage.backends.trivialbackend


    ## test if this is one of the yadage-specfic proxies
    if data['proxyname']=='CachedProxy':
        return CachedProxy.fromJSON(data,deserialization_opts)
    elif data['proxyname']=='TrivialProxy':
        return yadage.backends.trivialbackend.TrivialProxy.fromJSON(data)

    # it must be a packtivity proxy
    from_pack_proxy = packtivity.backendutils.load_proxy(
        data,
        deserialization_opts = deserialization_opts,
        best_effort_backend = False
    )
    if from_pack_proxy:
        return from_pack_proxy
    else:
        raise RuntimeError('unknown proxy found with name: {}'.format(data['proxyname']))



import copy
import itertools
import logging

import jq
import jsonpointer

import yadage.handlers.utils as utils
from packtivity.typedleafs import TypedLeafs

from ..stages import JsonStage
from ..tasks import packtivity_task
from ..utils import (get_init_spec, init_stage_spec, leaf_iterator_jsonlike,
                     outputReference, pointerize, process_jsonlike)
from .expression_handlers import handlers as exprhandlers

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()

# A scheduler does the following things:
#   - attached new nodes to the DAG
# - for each added step
#     - the step is given a name
#     - the step parameters are determined using the scheduler spec and context
#     - a list of used inputs (in the form of [stepname,outputkey,index])

def isExpression(parameter):
    return type(parameter)==dict and 'expression_type' in parameter

def select_parameter(wflowview, parameter):
    '''
    Evaluates parameter expressions (if needed) in the context of a workflow view

    :param wflowview: the workflow view on which to evaluete possible value expressions
    :param parameter: either a non-dict value or a JSON-like dict for a
                      supported value expression
    :return: the parameter value
    '''
    log.debug('selecting parameter %s', parameter)
    if isExpression(parameter):
        handler = exprhandlers[parameter['expression_type']]
        value = handler(wflowview, parameter)
    else:
        value = parameter
    return value

def finalize_value(wflowview, value):
    '''
    finalize a value by recursively resolving references

    :param wflowview: the workflow view against which to resolve upstream references
    :param step: the step for which to track usage of upstream references
    :param value: the parameter value. May be a output reference, or a JSON value type
    :return: finalized parameter value
    '''
    if type(value) == outputReference:
        v = wflowview.dag.getNode(value.stepid).result.resolve_ref(value.pointer)
        return finalize_value(wflowview, v)
    else:
        return value

def finalize_input(jsondata,wflowview):
    '''
    evaluate final values of step parameters by resolving
    references to a upstream output and contextualizing stateful
    parameters. Also tracks usage of upstream references for the step

    :param wflowview: the workflow view view against which to resolve any upstream references
    :param step: the step that for which to track usage of upstream references
    :param jsondata: the prospective step parameters

    :return: finalized step parameters
    '''

    result = copy.deepcopy(jsondata)
    inputs = []
    for leaf_pointer, leaf_value in leaf_iterator_jsonlike(jsondata):
        if type(leaf_value) == outputReference: inputs.append(leaf_value)
        leaf_pointer.set(result,finalize_value(wflowview, leaf_value))
    return result, inputs

def step_or_stages(name, spec, state_provider, inputs, parameters, dependencies):
    '''
    :param name: name of the eventual (init-)step
    :param spec: the stage spec
    :param state_provider: the stage's state provider
    :param dependencies: dependent nodes

    :return: yadage or init step object
    '''
    if 'step' in spec:
        depstates = set(d.task.state for d in dependencies if d.task.state)
        step_state = state_provider.new_state(name,depstates)
        p = packtivity_task(name=name, spec=spec['step'], state=step_state)
        p.s(**parameters)
        p.used_inputs(inputs)
        return p,None
    elif 'workflow' in spec:
        name = 'init_{}'.format(name)
        init_spec  = init_stage_spec(parameters.json(), discover = False, used_inputs=[x.json() for x in inputs], name = 'init', nodename = name)
        return None, [init_spec] + spec['workflow']['stages']
    elif 'cases' in spec:
        for x in spec['cases']:
            if parameters.jq(x['if']).json():
                log.info('selected case %s', x['if'])
                return step_or_stages(name,x, state_provider, inputs, parameters, dependencies)
        log.info('no case selected on pars %s', parameters)
        return None, None
    raise RuntimeError('do not know what kind of stage spec we are dealing with. %s', spec.keys())

def register_expressions(stage, expressions):
    if not expressions: return
    for key, expression in expressions.items():
        stage.view.addValue(key, expression)

def addStepOrWorkflow(name, stage, parameters, inputs, spec):
    '''
    adds a step or a sub-workflow based on a init step

    :param str name: the name of the step or sub-workflow
    :param stage: the stage from which to use state context and workflow view
    :param step: either a packtivity_task (for normal workflow steps) initstep object (for sub-workflows)
    :param spec: the stage spec

    :return: None
    '''
    dependencies = [stage.view.dag.getNode(k.stepid) for k in inputs]
    step,stages = step_or_stages(name,spec,stage.state_provider, inputs, parameters, dependencies)

    if step:
        stage.addStep(step)
        log.debug('scheduled a step')

    if stages: #subworkflow case
        depstates = [d.task.state for d in set(dependencies) if d.task.state]
        new_provider = stage.state_provider.new_provider(name, init_states = depstates)
        subrules = [JsonStage(s, new_provider) for s in stages]
        stage.addWorkflow(subrules,
            isolate = True
        )
        log.debug('scheduled a subworkflow')

def get_parameters(parameters):
    '''
    retrieve parameters from the spec

    :param spec: the stage spec
    :return: a JSON-like object of stage parameters
    '''
    return {x['key']: x['value']for x in parameters}

@scheduler('singlestep-stage')
def singlestep_stage(stage, spec):
    '''
    a simple state that adds a single step/workflow. The node is attached
    to the DAG based on used upstream outputs

    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    '''
    log.debug('scheduling singlestep stage with spec:\n%s', spec)
    parameters = {
        k: select_parameter(stage.view, v) for k, v in get_parameters(spec['parameters']).items()
    }
    finalized, inputs = finalize_input(parameters, stage.view)
    finalized = TypedLeafs(finalized, getattr(stage.state_provider,'datamodel',None))
    addStepOrWorkflow(stage.name, stage, finalized, inputs, spec)
    register_expressions(stage, spec.get('register_values'))

def chunk(alist, chunksize):
    '''split a list into equal-sized chunks of size chunksize'''
    return [alist[x:x+chunksize] for x in range(0, len(alist), chunksize)]


def partition(alist, partitionsize):
    '''split a list into partitionsize parts'''
    total_len = len(alist)
    if partitionsize > total_len:
        partitionsize = total_len
    assert partitionsize <= total_len
    end = 0
    partitioned = []
    for k in range(partitionsize):
        begin = end
        end =  end + (total_len+k)//partitionsize
        partitioned.append(alist[begin:end])
    return partitioned

def groupmany(iterable,batchsize = None, partitionsize = None):
    if batchsize:
        return chunk(list(iterable), batchsize)
    if partitionsize:
        return partition(list(iterable), partitionsize)
    return iterable


def scatter(parameters, scatter, batchsize = None, partitionsize = None):
    '''
    convert a parameter set and scatter definition into a list
    of single parameter sets.

    :param parameters: the parameter definition
    :param scatter: scattering method. One of 'zip' or 'cartesian'

    :return: list of parameter sets
    '''

    log.debug('scattering:  pars: %s scatter: %s batchsisze: %s partitionsize: %s',
        parameters, scatter, batchsize, partitionsize
    )
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter['parameters']:
        to_scatter[scatpar] = groupmany(commonpars.pop(scatpar), batchsize, partitionsize)

    singlesteppars = []
    if scatter['method'] == 'zip':
        keys, zippable = zip(*[(k, v) for k, v in to_scatter.items()])
        for zipped in zip(*zippable):
            individualpars = dict(zip(keys, zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter['method'] == 'cartesian':
        for what in itertools.product(*[to_scatter[k] for k in scatter['parameters']]):
            individualpars = dict(zip(scatter['parameters'], what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars


@scheduler('multistep-stage')
def multistep_stage(stage, spec):
    '''
    a stage that attaches an array of nodes to the DAG. The number of nodes
    is determined by a scattering recipe. Currently two algs are supported

    - ``zip``: one or more arrays of length n are iterated through in lock-step.
       n nodes are added to the DAG where the  parameters values are set to
           the values in the iteration
    - ``cartesian``:  a cartesian product of a number of arrays (possibly different sizes)
       adds n1 x n2 x ... nj nodes.

    Nodes are attached to the DAG based on used upstream inputs

    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    '''
    log.debug('scheduling multistep stage with spec:\n%s', spec)
    parameters = {
        k: select_parameter(stage.view, v) for k, v in get_parameters(spec['parameters']).items()
    }
    singlesteppars = scatter(parameters, spec['scatter'], spec.get('batchsize'), spec.get('partitionsize'))
    for i, pars in enumerate(singlesteppars):
        singlename = '{}_{}'.format(stage.name, i)
        finalized, inputs = finalize_input(pars, stage.view)
        finalized = TypedLeafs(finalized, getattr(stage.state_provider,'datamodel',None))
        addStepOrWorkflow(singlename, stage, finalized, inputs, spec)
    register_expressions(stage, spec.get('register_values'))


def process_noderef(leafobj,resultscript,view):
    n = view.dag.getNode(leafobj['_nodeid'])
    return jq.jq(resultscript).transform(pointerize(n.result,False,n.identifier), multiple_output = True)

def process_wflowref(leafobj,view):
    nodeselector, resultscript = leafobj['$wflowref']
    nodestruct = jq.jq(nodeselector).transform(view.steps, multiple_output = True)
    return process_jsonlike(nodestruct, 'has("_nodeid")', lambda x: process_noderef(x, resultscript, view))

def process_wflowpointer(leafobj):
    p = leafobj['$wflowpointer']
    return outputReference(p['step'],jsonpointer.JsonPointer(p['result']))

@scheduler('jq-stage')
def jq_stage(stage, spec):
    '''
    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    '''
    binds = spec['bindings']
    binds = process_jsonlike(binds,'has("$wflowref")',lambda x: process_wflowref(x,stage.view))
    log.info('transforming binds: %s', binds)
    stagescript = spec['stepscript']
    singlesteps = jq.jq(stagescript).transform(binds,multiple_output = False)

    singlesteppars = map(
        lambda x: process_jsonlike(x, 'has("$wflowpointer")',process_wflowpointer),
        singlesteps
    )

    postscript = spec['postscript']
    for i, pars in enumerate(singlesteppars):
        singlename = '{}_{}'.format(stage.name, i)

        finalized, inputs = finalize_input(pars, stage.view)
        log.info('postscripting: %s',finalized)
        after_post = jq.jq(postscript).transform(finalized,multiple_output = False)

        log.info('finalized to: %s',after_post)
        addStepOrWorkflow(singlename, stage, after_post, inputs, spec)
    register_expressions(stage, spec.get('register_values'))

@scheduler('init-stage')
def init_stage(stage, spec):
    '''
    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    '''
    inputs = []
    if spec.get('inputs'):
        inputs = map(outputReference.fromJSON, spec['inputs'])
        log.info('initializing scope from dependent tasks')
    else:
        log.info('initializing scope from dependent tasks')

    depstates = stage.state_provider.init_states if stage.state_provider else []

    if stage.state_provider:
        step_state = stage.state_provider.new_state(stage.name,depstates, readonly = True)
    else:
        step_state = None

    init_spec = get_init_spec(discover = spec['discover'])
    task = packtivity_task(spec['nodename'] or stage.name, init_spec, step_state)
    task.s(**spec['parameters'])
    task.used_inputs(inputs)
    stage.addStep(task)
    register_expressions(stage, spec.get('register_values'))



def handler_decorator():
    handlers = {}

    def decorator(name):
        def wrap(func):
            handlers[name] = func
            return func
        return wrap
    return handlers, decorator



import logging
import os
import shutil

from packtivity.statecontexts import load_state
from packtivity.statecontexts.posixfs_context import LocalFSState

from ..utils import prepare_workdir_from_archive

log = logging.getLogger(__name__)




def _merge_states(lhs,rhs):
    return LocalFSState(lhs.readwrite + rhs.readwrite,lhs.readonly + rhs.readonly)

class LocalFSProvider(object):
    def __init__(self, *base_states, **kwargs):
        base_states = list(base_states)
        self.nest = kwargs.get('nest', True)
        self.ensure = kwargs.get('ensure')
        self.init_states = kwargs.get('init_states')
        first = base_states.pop()
        assert first

        self.base = first

        while base_states:
            next_state = base_states.pop()
            if not next_state:
                continue
            self.base = _merge_states(self.base,next_state)

    def new_provider(self,name, init_states = None):
        new_base_ro = self.base.readwrite + self.base.readonly
        new_base_rw = [os.path.join(self.base.readwrite[0],name)]
        return LocalFSProvider(LocalFSState(new_base_rw,new_base_ro),
                               nest = self.nest,
                               ensure = self.ensure,
                               init_states = init_states or [])


    def new_state(self,name, dependencies, readonly = False):
        '''
        creates a new context from an existing context.

        if subdir is True it declares a new read-write nested under the old
        context's read-write and adds all read-write and read-only locations
        of the old context as read-only. This is recommended as it makes rolling
        back changes to the global state made in this context easy.

        else the same readwrite/readonly configuration as the parent context is used
        '''

        readwrites = []
        if not readonly:
            readwrites = ['{}/{}'.format(self.base.readwrite[0],name)] if self.nest else self.base.readwrite

        log.debug('new state is: rw: %s', readwrites)
        state_identifier = name.replace('/','_') # replace in case name is nested path
        newstate = LocalFSState(readwrite = readwrites, identifier = state_identifier, dependencies = dependencies)

        if not readonly and self.ensure:
            newstate.ensure()

        return newstate

    def json(self):
        return {
            'state_provider_type': 'localfs_provider',
            'base_state': self.base.json(),
            'init_states': [s.json() for s in self.init_states] if self.init_states else [],
            'nest': self.nest,
            'ensure': self.ensure
        }

    @classmethod
    def fromJSON(cls,jsondata, deserialization_opts):
        return cls(
            load_state(jsondata['base_state'],deserialization_opts),
            nest = jsondata['nest'],
            ensure = jsondata['ensure'],
            init_states = [load_state(x,deserialization_opts) for x in jsondata['init_states']]
        )

def setup_provider(dataarg, dataopts):
    workdir = dataarg.split(':',2)[1]
    read = dataopts.get('read',None)
    nest = dataopts.get('nest',True)
    ensure = dataopts.get('ensure',True)
    overwrite = dataopts.get('overwrite',False)

    if overwrite and os.path.exists(workdir):
        shutil.rmtree(workdir)

    init_states = []

    initdir = os.path.join(workdir,dataopts.get('initdir','init'))
    inputarchive = dataopts.get('inputarchive',None)
    if initdir:
        if inputarchive:
            prepare_workdir_from_archive(initdir, inputarchive)
        init_state = LocalFSState(readonly = [initdir])
        init_states.append(init_state)
    writable_state = LocalFSState([workdir])
    return LocalFSProvider(read,writable_state, ensure = ensure, nest = nest, init_states = init_states)



from packtivity.asyncbackends import PacktivityProxyBase
from packtivity.typedleafs import TypedLeafs


class TrivialProxy(PacktivityProxyBase):
    '''
    A trivial proxy that carries the results and status already
    '''

    def __init__(self, status, resultdata, datamodel):
        self.status = status
        self.resultdata = resultdata
        self.datamodel = datamodel

    def proxyname(self):
        return 'TrivialProxy'

    def details(self):
        return {
            'resultdata':self.resultdata,
            'datamodel': self.datamodel,
            'status':self.status
        }

    @classmethod
    def fromJSON(cls,data):
        return cls(**data['proxydetails'])

class TrivialBackend(object):
    '''
    A trivial backend that can only return proxy based information, no submission
    '''

    def submit(self, task):
        raise NotImplementedError(
            'The trivial proxy is not made for submission')

    def result(self, resultproxy):
        return TypedLeafs(resultproxy.resultdata, resultproxy.datamodel)

    def expected_result(self, resultproxy):
        return None

    def ready(self, resultproxy):
        # when we have a proxy it is by definition ready...
        return True

    def successful(self, resultproxy):
        return resultproxy.status == 'SUCCESS'

    def fail_info(self, resultproxy):
        return None



from adage.pollingexec import yes_man

from .interactive import custom_decider, decide_step, interactive_deciders
from .utils import advance_coroutine

strategies = {}
def strategy(name):
    def wrap(func):
        strategies[name] = func
        return func
    return wrap

def get_strategy(name):
    return strategies[name]()

@strategy('interactive')
def interactive_strategy():
    extend, submit = interactive_deciders()
    return dict(
            extend_decider = extend,
            submit_decider = submit
    )

@strategy('askforsubmit')
def askforsubmit():
    extend_decider = yes_man()
    advance_coroutine(extend_decider)

    submit_decider = custom_decider(decide_step, idbased = False)()
    advance_coroutine(submit_decider)

    return dict(
            extend_decider = extend_decider,
            submit_decider = submit_decider
    )



import logging

from .steering_object import YadageSteering
from .utils import setupbackend_fromstring
from .wflowstate import make_deserializer

log = logging.getLogger(__name__)

def connect(metadir, accept_metadir, ctrlstring, ctrlopts, modelsetup, modelopts, backendstring):
    ys = YadageSteering.connect(
        accept_metadir = accept_metadir,
        metadir = metadir,
        ctrlstring = ctrlstring,
        ctrlopts = ctrlopts,
        modelsetup = modelsetup,
        modelopts = modelopts
    )
    if backendstring:
        ys.controller.backend = setupbackend_fromstring(backendstring)
    return ys

def preview_rule(wflow, name = None, identifier=None):
    stateopts = {}
    wflowmaker = make_deserializer(stateopts)
    newflow = wflowmaker(wflow.json())

    if identifier:
        rule = newflow.view().getRule(identifier=identifier)
    else:
        offset, name = name.split('/')
        rule = newflow.view(offset).getRule(name)

    if not rule.applicable(newflow):
        log.warning('rule not applicable')
        return

    rule.apply(newflow)
    newflow.rules.remove(rule)
    newflow.applied_rules.append(rule)

    existing_rules = [x.identifier for x in (wflow.rules + wflow.applied_rules)]
    existing_nodes = wflow.dag.nodes()

    new_rules = [{'name': x.rule.name, 'offset': x.offset} for x in newflow.rules if x.identifier not in existing_rules]
    new_nodes = [{'name': newflow.dag.getNode(n).name, 'parents': newflow.dag.predecessors(n)} for n in newflow.dag.nodes() if n not in existing_nodes]
    return new_rules, new_nodes



import contextlib
import importlib
import logging

from adage.wflowcontroller import BaseController

from .reset import collective_downstream, remove_rules, reset_steps, undo_rules
from .wflow import YadageWorkflow

log = logging.getLogger(__name__)

def setup_controller(model = None, controller = 'frommodel', ctrlopts = None):
    '''
    return controller instance based on state configuration. For
    transaction-based states, returns PersistentController, for in-
    memory states returns in BaseController
    '''

    ctrlopts  = ctrlopts or {}

    if controller == 'frommodel':
        if isinstance(model, YadageWorkflow):
            return BaseController(model)
        else:
            return PersistentController(model)
    if controller.startswith('http'):
        try:
            from yadagehttpctrl.clientcontroller import YadageHTTPController
            ctrl = YadageHTTPController(server = controller, **ctrlopts)
            return ctrl
        except ImportError:
            log.exception('try installing yadagehttpctrl')
    if controller.startswith('py:'):
        _, module, ctrlclass = controller.split(':')
        module = importlib.import_module(module)
        ctrlclass = getattr(module,ctrlclass)
        if ctrlopts.pop('pass_model',False):
            ctrlopts['model'] = model
        return ctrlclass(**ctrlopts)
    else:
        raise RuntimeError('unknown controller type %s', controller)

class PersistentController(BaseController):
    '''
    workflow controller, that explicltly calls transaction methods on non read-only operations on the workflow state
    '''
    def __init__(self, model, backend = None):
        '''
        :param model: the model on whih the controller will operate
        :param backend: the backend to against which to check workflow state.

        :return: the controller instance
        '''
        self.model = model
        super(PersistentController, self).__init__(self.model.load(),backend)


    @contextlib.contextmanager
    def transaction(self, sync = True):
        '''the transaction context. will commit model to persistent store on exit.'''
        self.adageobj = self.model.load()
        if sync: self.sync_backend()
        yield

        isvalid = self.validate()
        if not isvalid:
            #raise RuntimeError('was about to commit invalid data!')
            log.warning('commit is in valid %s', isvalid)
        self.model.commit(self.adageobj)

    def submit_nodes(self, nodeids):
        '''
        submit nodes to backend

        :param nodeids: list of ids of nodes to be submitted
        :return: None
        '''
        with self.transaction():
            nodes = [self.adageobj.dag.getNode(nodeid) for nodeid in nodeids]
            # log.info('submitting nodes to backend: %s', nodes)
            super(PersistentController,self).submit_nodes(nodes)

    def apply_rules(self, ruleids):
        '''
        apply rules to workflow instance

        :param ruleids: list of ids of reuls to be applied
        :return: None
        '''
        with self.transaction():
            rules = [r for r in self.adageobj.rules if r.identifier in ruleids]
            super(PersistentController,self).apply_rules(rules)

    def sync_backend(self):
        '''
        synchronize node data with backend
        '''
        with self.transaction(sync = False): # disable sync to avoid infinite recursion
            super(PersistentController,self).sync_backend()

    def applicable_rules(self):
        '''
        :return: a list of rules whose predicate is fulfilled
        '''
        applicable_rules = super(PersistentController,self).applicable_rules()
        return [x.identifier for x in applicable_rules]

    def submittable_nodes(self):
        '''
        :return: a list of nodes with sucessfull and completed upstream
        '''
        submittable_nodes = super(PersistentController,self).submittable_nodes()
        return [x.identifier for x in submittable_nodes]

    def add_rules(self, rulespecs, dataarg, offset = '', groupname = None , dataopts = None):
        log.info('adding %s rules', len(rulespecs))
        from .stages import JsonStage
        from .utils  import state_provider_from_string
        sp = state_provider_from_string(dataarg,dataopts)
        rules = [JsonStage(json, sp) for json in rulespecs]
        with self.transaction():
            self.adageobj.view(offset).addWorkflow(rules,groupname)

    def patch_rule(self, ruleid, patchspec):
        with self.transaction():
            rule = [x for x in self.adageobj.rules if x.identifier == ruleid][0]
            rule.rule.stagespec = patchspec

    def undo_rules(self, ruleids):
        '''
        :param nodeids: list of ids of nodes to reset
        :return: None
        '''
        with self.transaction():
            undo_rules(self.adageobj, ruleids)

    def remove_rules(self, ruleids):
        with self.transaction():
            remove_rules(self.adageobj, ruleids)

    def reset_nodes(self, nodeids):
        '''
        :param nodeids: list of ids of nodes to reset
        :return: None
        '''
        to_reset = nodeids + collective_downstream(self.adageobj, nodeids)
        with self.transaction():
            reset_steps(self.adageobj,to_reset)



import adage
from adage.serialize import obj_to_json

from .stages import JsonStage, OffsetStage
from .wflowview import WorkflowView


def json_or_nil(x):
    return None if x is None else x.json()

class YadageWorkflow(adage.adageobject):
    '''
    The overall workflow state object that extends the basic
    Adage state object by two bookkeeping structures.
    '''

    def __init__(self):
        super(YadageWorkflow, self).__init__()
        self.stepsbystage = {}
        self.bookkeeping = {}
        self.values = {}

    def view(self, offset=''):
        return WorkflowView(self, offset)

    def json(self):
        data = obj_to_json(self,
                           ruleserializer=json_or_nil,
                           taskserializer=json_or_nil,
                           proxyserializer=json_or_nil,
                           )
        data['bookkeeping'] = self.bookkeeping
        data['stepsbystage'] = self.stepsbystage
        data['values'] = self.values
        return data

    @classmethod
    def fromJSON(cls, data,
            proxydeserializer = lambda data: None,
            state_provider_deserializer = lambda data: None,
            task_deserializer = lambda data: None,
            backend=None):
        instance = cls()
        instance.rules = [OffsetStage.fromJSON(x,state_provider_deserializer) for x in data['rules'] ]
        instance.applied_rules = [OffsetStage.fromJSON(x,state_provider_deserializer) for x in data['applied'] ]
        instance.bookkeeping = data['bookkeeping']
        instance.stepsbystage = data['stepsbystage']
        instance.values = data['values']

        instance.dag = adage.serialize.dag_from_json(
            data['dag'],
            task_deserializer,
            proxydeserializer,
            backend
        )
        return instance

    @classmethod
    def createFromJSON(cls, jsondata, state_provider):
        instance = cls()
        rules = [JsonStage(stagedata, state_provider) for stagedata in jsondata['stages']]
        instance.view().addWorkflow(rules)
        return instance



import json
import logging
import os

import adage
import adage.backends
import yadage.serialize as serialize
import yadage.utils as utils
import yadage.workflow_loader as workflow_loader
import yadageschemas

from .controllers import setup_controller
from .utils import setupbackend_fromstring
from .wflow import YadageWorkflow
from .wflowstate import load_model_fromstring

log = logging.getLogger(__name__)

class YadageSteering(object):
    '''
    high level steering object to manage worklfow execution
    '''
    def __init__(self,loggername = __name__):
        self.log = logging.getLogger(loggername)
        self.metadir = None
        self.controller = None
        self.rootprovider = None
        self.adage_kwargs = {}

    @classmethod
    def connect(cls, metadir, ctrlstring, ctrlopts = None, modelsetup = None, modelopts = None, accept_metadir = False):
        model = None
        instance = cls()
        instance.prepare_meta(metadir, accept = accept_metadir)
        if modelsetup:
            model = load_model_fromstring(modelsetup,modelopts)
        instance.controller = setup_controller(
            model = model,
            controller = ctrlstring, ctrlopts = ctrlopts,
        )
        log.info('connected to model')
        return instance

    @classmethod
    def create(
        cls,
        dataarg = None,
        dataopts = None,
        workflow_json = None,
        workflow = None,
        toplevel = os.getcwd(),
        schemadir = yadageschemas.schemadir,
        validate=True,
        initdata = None,
        controller = 'frommodel',
        ctrlopts = None,
        metadir = None,
        accept_metadir = False,
        modelsetup = 'inmem',
        modelopts = None):

        ys = cls()
        if workflow_json:
            wflow_kwargs = dict(
                workflow_json = workflow_json
            )
        elif workflow:
            wflow_kwargs = dict(
                workflow = workflow, toplevel = toplevel,
                validate = validate, schemadir = schemadir
            )
        else:
            raise RuntimeError('need to initialize either from full JSON or remote location')

        ys.prepare(
            dataarg = dataarg, dataopts = dataopts,
            metadir = metadir, accept_metadir = accept_metadir,
        )

        ys.init_workflow(
            initdata = initdata,
            modelsetup = modelsetup,
            modelopts = modelopts,
            controller = controller,
            ctrlopts = ctrlopts,
            **wflow_kwargs
        )
        return ys



    @property
    def workflow(self):
        '''
        :return: the workflow object (from the controller)
        '''
        return self.controller.adageobj

    def prepare_meta(self,metadir = None, accept=False):
        '''
        prepare workflow meta-data directory

        :param metadir: the meta-data directory name
        :param accept: whether to accept an existing metadata directory
        '''
        self.metadir = self.metadir or metadir #maybe it's already set
        assert self.metadir
        if os.path.exists(self.metadir):
            if not accept:
                raise RuntimeError('yadage meta directory %s exists. explicitly accept', self.metadir)
        else:
            os.makedirs(self.metadir)
        self.adage_argument(workdir = os.path.join(self.metadir,'adage'))

    def prepare(self, dataarg, dataopts = None, accept_metadir = False, metadir = None):
        '''
        prepares workflow data state, with  possible initialization and sets up stateprovider used for workflow stages.
        if initialization data is provided

        sets self.metadir and self.rootprovider

        :param dataarg: mandatory state provider setup. generally <datatype>:<argument>. For
        :param dataopts: optional settings for state provider
        :param accept_metadir:
        :param metadir: meta-data directory
        '''

        self.metadir = metadir
        self.rootprovider = utils.state_provider_from_string(dataarg, dataopts)
        dataopts = dataopts or {}
        is_local_data = len(dataarg.split(':',1)) == 1
        if is_local_data:
            self.metadir = self.metadir or '{}/_yadage/'.format(dataarg)
        else:
            assert self.metadir #for non-default provider, require metadir to be set
        self.prepare_meta(accept = accept_metadir)

    def init_workflow(self,
                      workflow = None,
                      initdata = None,
                      toplevel = os.getcwd(),
                      workflow_json = None,
                      modelsetup = 'inmem',
                      modelopts = None,
                      controller = 'frommodel',
                      ctrlopts = None,
                      validate = True,
                      schemadir = yadageschemas.schemadir):
        '''
        load workflow from spec and initialize it

        :param workflow: the workflow spec source
        :param toplevel: base URI against which to resolve JSON references in the spec
        :param initdata: initialization data for workflow

        prepares initial workflow object and sets self.controller
        '''

        if not self.rootprovider:
            raise RuntimeError('need to setup root state provider first. run .prepare() first')

        if not workflow_json and not workflow:
            raise RuntimeError('need to provide either direct workflow spec or source to load from')

        if workflow_json:
            if validate: workflow_loader.validate(workflow_json)
        else:
            workflow_json = workflow_loader.workflow(
                workflow,
                toplevel=toplevel,
                schemadir=schemadir,
                validate=validate
            )

        with open('{}/yadage_template.json'.format(self.metadir), 'w') as f:
            json.dump(workflow_json, f)
        workflowobj = YadageWorkflow.createFromJSON(workflow_json, self.rootprovider)
        if initdata:
            log.info('initializing workflow with %s',initdata)
            workflowobj.view().init(initdata, self.rootprovider, discover = True)
        else:
            log.info('no initialization data')

        model = load_model_fromstring(modelsetup,modelopts,workflowobj)
        self.controller = setup_controller(
            model = model,
            controller = controller, ctrlopts = ctrlopts,
        )

    def adage_argument(self,**kwargs):
        '''
        add keyword arguments for workflow execution (adage)

        :param kwargs: adage keyword arguments (see adage documentation for options)
        '''
        self.adage_kwargs.update(**kwargs)

    def run_adage(self, backend = 'auto', **adage_kwargs):
        '''
        execution workflow with adage based against given backend
        :param backend: backend to use for packtivity processing.
        '''
        if backend=='auto':
            #respect if the controller already has a backend wired up
            backend = self.controller.backend or setupbackend_fromstring('multiproc:auto')
            log.info('backend automatically set to %s', backend)
        elif backend:
            self.controller.backend = backend
        self.adage_argument(**adage_kwargs)
        adage.rundag(controller = self.controller, **self.adage_kwargs)

    def serialize(self):
        '''
        serialized workflow and backend states (stored in meta directory)
        '''
        serialize.snapshot(
            self.workflow,
            '{}/yadage_snapshot_workflow.json'.format(self.metadir),
            '{}/yadage_snapshot_backend.json'.format(self.metadir)
        )

    def visualize(self):
        '''
        generate workflow visualization (stored in meta directory)
        '''
        import yadage.visualize as visualize
        visualize.write_prov_graph(self.metadir, self.workflow, vizformat='png')
        visualize.write_prov_graph(self.metadir, self.workflow, vizformat='pdf')



import click

from .utils import advance_coroutine


def decide_rule(rule, controller, idbased):
    click.secho('we could extend DAG with rule', fg='blue')
    if idbased:
        rule = next(x for x in controller.adageobj.rules if x.identifier == rule)

    click.secho('rule: {}/{} ({})'.format(rule.offset,
                                          rule.rule.name, rule.identifier))
    # IPython.embed()
    try:
        resp = raw_input(click.style("Shall we? (y/N) ", fg='blue'))
    except NameError:
        resp = input(click.style("Shall we? (y/N) ", fg='blue'))

    shall = resp.lower() == 'y'
    if shall:
        click.secho('ok we will extend.', fg='green')
    else:
        click.secho('maybe another time... your response: {}'.format(
            resp), fg='yellow')
    return shall


def decide_step(node, controller, idbased):


    if idbased:
        node = controller.adageobj.dag.getNode(node)

    click.echo('we could submit a DAG node {}'.format(node))
    # IPython.embed()
    try:
        resp = raw_input(click.style("Shall we? (y/N) ", fg='magenta'))
    except NameError:
        resp = input(click.style("Shall we? (y/N) ", fg='magenta'))

    shall = resp.lower() == 'y'
    if shall:
        click.secho('ok we will submit.', fg='green')
    else:
        click.secho('will not submit for now... your response: {}'.format(
            resp), fg='yellow')
    return shall


def custom_decider(decide_func, idbased):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = yield decide_func(*data, idbased = idbased)
    return decider



def interactive_deciders(idbased = False):
    '''
    returns a tuple (extend,submit) of already-primed deciders for both
    extension and submission
    '''
    extend_decider = custom_decider(decide_rule, idbased)()
    advance_coroutine(extend_decider)  # prime decider

    submit_decider = custom_decider(decide_step, idbased)()
    advance_coroutine(submit_decider)  # prime decider

    return extend_decider, submit_decider



import os
import importlib
import logging
from .localposix import LocalFSProvider

log = logging.getLogger(__name__)

def load_provider(jsondata,deserialization_opts = None):

    log.debug('load_provider opts %s', deserialization_opts)
    deserialization_opts = deserialization_opts or {}
    if jsondata == None:
        return None
    if 'state_provider' in deserialization_opts:
        providerstring = deserialization_opts.get('state_provider','')
        if providerstring.startswith('py:'):
            _, module, providerclass = providerstring.split(':')
            module = importlib.import_module(module)
            providerclass = getattr(module,providerclass)
            provideropts = {}
            return providerclass.fromJSON(jsondata,**provideropts)
    if 'YADAGE_STATEPROVIDER' in os.environ:
        module = importlib.import_module(os.environ['YADAGE_STATEPROVIDER'])
        return module.load_provider(jsondata)
    if jsondata['state_provider_type'] == 'localfs_provider':
        return LocalFSProvider.fromJSON(jsondata, deserialization_opts)
    raise TypeError('unknown provider type {}'.format(jsondata['state_provider_type']))



import os
from setuptools import setup, find_packages

deps = [
    'adage',
    'packtivity',
    'yadage-schemas',
    'click',
    'psutil',
    'requests[security]>=2.9',
    'pyyaml',
    'jsonref',
    'jsonschema',
    'jsonpointer>=1.10',
    'jsonpath_rw',
    'checksumdir',
    'glob2',
    'jq'
]


if 'READTHEDOCS' in os.environ:
  deps = [
    'yadage-schemas',
    'click',
    'psutil',
    'requests[security]>=2.9',
    'pyyaml',
    'jsonref',
    'jsonschema',
    'jsonpointer>=1.10',
    'jsonpath_rw',
    'checksumdir',
    'glob2',
  ]

setup(
  name = 'yadage',
  version = '0.15.2',
  description = 'yadage - YAML based adage',
  url = '',
  author = 'Lukas Heinrich',
  author_email = 'lukas.heinrich@cern.ch',
  packages = find_packages(),
  include_package_data = True,
  install_requires = deps,
  extras_require = {
    'celery' : [
       'celery',
       'redis'
    ],
    'viz': [
        #manually adding extras of adage[extra] because of pip
        #issue https://github.com/pypa/pip/issues/3189
        'pydot2',
        'pygraphviz',
        'pydotplus'
    ],
    'develop': [
       'pyflakes',
       'isort',
       'pytest>=3.2.0',
       'pytest-cov>=2.5.1',
       'python-coveralls'
    ]
  },
  entry_points = {
      'console_scripts': [
          'yadage-run=yadage.steering:main',
          'yadage-manual=yadage.manualcli:mancli',
          'yadage-util=yadage.utilcli:utilcli',
      ],
  },
  dependency_links = [
  ]
)
